"""
Application of QtDraw.

This module provides application of QtDraw.
"""

import os
import warnings
from pathlib import Path
import logging
from PySide6.QtWidgets import QWidget, QMessageBox, QFileDialog, QDialog
from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from qtdraw.widget.custom_widget import Label, Layout, LineEdit, HBar, Button, Combo, VSpacer
from qtdraw.core.pyvista_widget import PyVistaWidget, Window
from qtdraw.util.logging_util import LogWidget
from qtdraw.core.pyvista_widget_setting import widget_detail as detail
from qtdraw.core.dialog_preference import PreferenceDialog
from qtdraw.core.dialog_about import AboutDialog
from qtdraw.util.util import check_multipie, create_style_sheet
from qtdraw.core.dialog_about import get_version_info


# ==================================================
class QtDraw(Window):
    # ==================================================
    def __init__(self, filename=None, status=None, preference=None):
        """
        3D drawing tool.

        Args:
            filename (str, optional): filename.
            status (dict, optional): status.
            preference (dict, optional): preference.
        """
        log_id = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critial": logging.CRITICAL,
        }
        level = detail["log_level"]
        super().__init__(log_id[level])
        self.debug = level == "debug"

        warnings.showwarning = lambda message, category, filename, lineno, file=None, line=None: logging.warning(str(message))
        warnings.filterwarnings("default", category=DeprecationWarning)

        self.pref_dialog = None  # preference dialog.
        self.multipie_dialog = None  # MultiPie dialog.
        if self.debug:
            self.actor_dialog = None  #  actor list dialog.
            self.data_dialog = None  # raw data dialog.
            self.status_dialog = None  # status data dialog.
            self.pref_data_dialog = None  # preference data dialog.
            self.camera_dialog = None  # camera dialog.

        self.create_gui()
        self.pyvista_widget.set_property(status, preference)
        self._update_panel()
        self.create_connection()

        if filename is not None and os.path.exists(filename):
            self.load_file(filename)

        # event loop.
        self.show()

    # ==================================================
    def create_gui(self):
        """
        create gui.

        :meta private:
        """
        self.resize(1000, 500)

        # central grid.
        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)
        layout = Layout(central_widget)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setVerticalSpacing(5)

        # PyVista widget.
        self.info_dialog = LogWidget("Information", None)
        txt = get_version_info()
        self.write_info(txt)
        self.pyvista_widget = PyVistaWidget(self)

        # Control panel.
        self.panel = self.create_panel(parent=self)

        # Status bar.
        self.status = Label(self, "")
        self.status.setAlignment(Qt.AlignRight)
        self.status.setContentsMargins(0, 0, 10, 0)
        self.status.setText(self.pyvista_widget.copyright)

        layout.addWidget(self.pyvista_widget, 0, 0, 1, 1)
        layout.addWidget(self.panel, 0, 1, 1, 1)
        layout.addWidget(self.status, 1, 0, 1, 2)

    # ==================================================
    def open_file(self):
        """
        Open file dialog.

        :meta private:
        """
        ext = detail["extension"]
        mat = "*" + " *".join(detail["ext_material"])
        cwd = os.getcwd()
        ext_set = f"QtDraw, CIF, VESTA, XSF Files (*{ext} {mat})"
        filename, _ = QFileDialog.getOpenFileName(self, "Open File", cwd, ext_set, options=QFileDialog.Options())
        self.sender().setDown(False)  # reset push button.

        if filename:
            filename = Path(filename)
            cur_ext = str(filename.suffix)
            if cur_ext == "":
                filename = str(filename) + ext
            if ext in ext_set:
                self.load_file(str(filename))

    # ==================================================
    def load_file(self, filename):
        """
        Load file.

        Args:
            filename (str): full file name.

        :meta private:
        """
        self.pyvista_widget.load(filename)

        # to avoid redraw object twice.
        # disconnect unit cell.
        self.uc_combo_crystal.currentTextChanged.disconnect(self._set_crystal)
        self.uc_edit_origin.returnPressed.disconnect(self._set_origin)
        self.uc_edit_a.returnPressed.disconnect(self._set_a)
        self.uc_edit_b.returnPressed.disconnect(self._set_b)
        self.uc_edit_c.returnPressed.disconnect(self._set_c)
        self.uc_edit_alpha.returnPressed.disconnect(self._set_alpha)
        self.uc_edit_beta.returnPressed.disconnect(self._set_beta)
        self.uc_edit_gamma.returnPressed.disconnect(self._set_gamma)
        # disconnect view.
        self.view_button_clip.toggled.disconnect(self._set_clip)
        self.view_button_repeat.toggled.disconnect(self._set_repeat)
        self.view_button_nonrepeat.pressed.disconnect(self._nonrepeat)
        self.view_edit_lower.returnPressed.disconnect(self._set_lower)
        self.view_edit_upper.returnPressed.disconnect(self._set_upper)

        self._update_panel()

        # reconnect unit cell.
        self.uc_combo_crystal.currentTextChanged.connect(self._set_crystal)
        self.uc_edit_origin.returnPressed.connect(self._set_origin)
        self.uc_edit_a.returnPressed.connect(self._set_a)
        self.uc_edit_b.returnPressed.connect(self._set_b)
        self.uc_edit_c.returnPressed.connect(self._set_c)
        self.uc_edit_alpha.returnPressed.connect(self._set_alpha)
        self.uc_edit_beta.returnPressed.connect(self._set_beta)
        self.uc_edit_gamma.returnPressed.connect(self._set_gamma)
        # reconnect view.
        self.view_button_clip.toggled.connect(self._set_clip)
        self.view_button_repeat.toggled.connect(self._set_repeat)
        self.view_button_nonrepeat.pressed.connect(self._nonrepeat)
        self.view_edit_lower.returnPressed.connect(self._set_lower)
        self.view_edit_upper.returnPressed.connect(self._set_upper)

    # ==================================================
    def save_file(self):
        """
        Save file dialog.

        :meta private:
        """
        ext = detail["extension"]
        file = Path.cwd() / (self.pyvista_widget._status["model"] + ext)
        ext_set = f"QtDraw Files (*{ext})"
        filename, _ = QFileDialog.getSaveFileName(self, "Save File", str(file.name), ext_set, options=QFileDialog.Options())
        self.sender().setDown(False)  # reset push button.

        if filename:
            filename = Path(filename)
            cur_ext = filename.suffix
            if cur_ext == "":
                filename = filename / ext
            if cur_ext == ext:
                self.pyvista_widget.save(str(filename))
                self._update_title()

    # ==================================================
    def _save_screenshot(self):
        """
        Save screenshot dialog.

        :meta private:
        """
        ifile = " ".join(detail["image_file"]).replace(".", "*.")
        gfile = " ".join(detail["vector_file"]).replace(".", "*.")
        ext = detail["image_file"][0]

        file = Path.cwd() / (self.pyvista_widget._status["model"] + ext)

        ext_set = f"Image Files ({ifile});;Graphic Files ({gfile})"
        filename, _ = QFileDialog.getSaveFileName(self, "Save Screenshot", str(file.name), ext_set)
        self.sender().setDown(False)  # reset push button.

        self.pyvista_widget.save_screenshot(filename)

    # ==================================================
    def create_panel(self, parent):
        """
        create right panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- right panel.

        :meta private:
        """
        panel = QWidget(parent)
        panel.setGeometry(60, 50, 180, 694)
        panel.setMaximumWidth(220)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 5, 5)

        uc = self.create_gui_unit_cell(panel)
        view = self.create_gui_view(panel)
        dataset = self.create_gui_dataset(panel)
        misc = self.create_gui_misc(panel)
        if self.debug:
            debug = self.create_gui_debug(panel)

        layout.addWidget(uc, 0, 0, 1, 1)
        layout.addWidget(HBar(), 1, 0, 1, 1)
        layout.addWidget(view, 2, 0, 1, 1)
        layout.addWidget(HBar(), 3, 0, 1, 1)
        layout.addWidget(dataset, 4, 0, 1, 1)
        layout.addWidget(HBar(), 5, 0, 1, 1)
        layout.addWidget(misc, 6, 0, 1, 1)
        if self.debug:
            layout.addWidget(HBar(), 7, 0, 1, 1)
            layout.addWidget(debug, 8, 0, 1, 1)
            layout.addItem(VSpacer(), 9, 0, 1, 1)
        else:
            layout.addItem(VSpacer(), 7, 0, 1, 1)

        return panel

    # ==================================================
    def create_gui_unit_cell(self, parent):
        """
        Create unit cell panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- unit cell panel.

        :meta private:
        """
        panel = QWidget(parent)
        layout = Layout(panel)

        label_uc = Label(parent, "UnitCell", True)

        label_crystal = Label(parent, "Crystal")
        self.uc_combo_crystal = Combo(
            parent, ["", "triclinic", "monoclinic", "orthorhombic", "tetragonal", "trigonal", "hexagonal", "cubic"]
        )

        label_origin = Label(parent, "Origin")
        self.uc_edit_origin = LineEdit(parent, "", ("list", ((3,), [""], 2)))

        self.uc_label_volume = Label(parent)

        label_a = Label(parent, "a")
        self.uc_edit_a = LineEdit(parent, "", ("sympy_float", 4))

        label_b = Label(parent, "b")
        self.uc_edit_b = LineEdit(parent, "", ("sympy_float", 4))

        label_c = Label(parent, "c")
        self.uc_edit_c = LineEdit(parent, "", ("sympy_float", 4))

        label_alpha = Label(parent, "\u03b1")
        self.uc_edit_alpha = LineEdit(parent, "", ("sympy_float", 2))

        label_beta = Label(parent, "\u03b2")
        self.uc_edit_beta = LineEdit(parent, "", ("sympy_float", 2))

        label_gamma = Label(parent, "\u03b3")
        self.uc_edit_gamma = LineEdit(parent, "", ("sympy_float", 2))

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.addWidget(label_uc, 0, 0, 1, 4)
        layout1.addWidget(self.uc_label_volume, 0, 4, 1, 4)

        panel2 = QWidget(parent)
        layout2 = Layout(panel2)
        layout2.setContentsMargins(0, 10, 0, 10)
        layout2.setVerticalSpacing(2)
        layout2.addWidget(label_a, 0, 0, 1, 1)
        layout2.addWidget(self.uc_edit_a, 0, 1, 1, 3)
        layout2.addWidget(label_alpha, 0, 4, 1, 1)
        layout2.addWidget(self.uc_edit_alpha, 0, 5, 1, 3)
        layout2.addWidget(label_b, 1, 0, 1, 1)
        layout2.addWidget(self.uc_edit_b, 1, 1, 1, 3)
        layout2.addWidget(label_beta, 1, 4, 1, 1)
        layout2.addWidget(self.uc_edit_beta, 1, 5, 1, 3)
        layout2.addWidget(label_c, 2, 0, 1, 1)
        layout2.addWidget(self.uc_edit_c, 2, 1, 1, 3)
        layout2.addWidget(label_gamma, 2, 4, 1, 1)
        layout2.addWidget(self.uc_edit_gamma, 2, 5, 1, 3)

        panel3 = QWidget(parent)
        layout3 = Layout(panel3)
        layout3.addWidget(label_crystal, 0, 0, 1, 2)
        layout3.addWidget(self.uc_combo_crystal, 0, 2, 1, 6)
        layout3.addWidget(label_origin, 1, 0, 1, 2)
        layout3.addWidget(self.uc_edit_origin, 1, 2, 1, 6)

        layout.addWidget(panel1, 0, 0, 1, 1)
        layout.addWidget(panel2, 1, 0, 1, 1)
        layout.addWidget(panel3, 2, 0, 1, 1)

        return panel

    # ==================================================
    def create_gui_view(self, parent):
        """
        Create view panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- view panel.

        :meta private:
        """
        panel = QWidget(parent)
        layout = Layout(panel)

        label_view = Label(parent, "View", True)
        self.view_button_clip = Button(parent, "clip", True)
        self.view_button_repeat = Button(parent, "repeat", True)
        self.view_button_nonrepeat = Button(parent, "non-repeat")

        label_lower = Label(parent, "lower")
        self.view_edit_lower = LineEdit(parent, "", ("list", ((3,), [""], 2)))
        label_upper = Label(parent, "upper")
        self.view_edit_upper = LineEdit(parent, "", ("list", ((3,), [""], 2)))

        self.view_button_x = Button(parent, "+x")
        self.view_button_y = Button(parent, "+y")
        self.view_button_z = Button(parent, "+z")
        self.view_button_xm = Button(parent, "-x")
        self.view_button_ym = Button(parent, "-y")
        self.view_button_zm = Button(parent, "-z")
        label_a = Label(parent, "a")
        label_b = Label(parent, "b")
        label_c = Label(parent, "c")
        self.view_combo_a = Combo(parent, [str(i) for i in range(-9, 10)])
        self.view_combo_b = Combo(parent, [str(i) for i in range(-9, 10)])
        self.view_combo_c = Combo(parent, [str(i) for i in range(-9, 10)])

        self.view_button_default = Button(parent, "default")
        self.view_button_bar = Button(parent, "bar", True)
        self.view_button_parallel = Button(parent, "parallel", True)
        self.view_button_grid = Button(parent, "grid", True)

        label_axis = Label(parent, "axis")
        self.view_combo_axis = Combo(parent, ["on", "axis", "full", "off"])
        label_cell = Label(parent, "cell")
        self.view_combo_cell = Combo(parent, ["single", "all", "off"])

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.addWidget(label_view, 0, 0, 1, 4)
        layout1.addWidget(self.view_button_nonrepeat, 0, 4, 1, 4)
        layout1.addWidget(self.view_button_clip, 1, 0, 1, 4)
        layout1.addWidget(self.view_button_repeat, 1, 4, 1, 4)

        panel2 = QWidget(parent)
        layout2 = Layout(panel2)
        layout2.setVerticalSpacing(2)
        layout2.addWidget(label_lower, 0, 0, 1, 2)
        layout2.addWidget(self.view_edit_lower, 0, 2, 1, 6)
        layout2.addWidget(label_upper, 1, 0, 1, 2)
        layout2.addWidget(self.view_edit_upper, 1, 2, 1, 6)

        panel3 = QWidget(parent)
        layout3 = Layout(panel3)
        layout3.setContentsMargins(0, 10, 0, 10)
        layout3.setVerticalSpacing(2)
        layout3.addWidget(self.view_button_default, 0, 0, 1, 3)
        layout3.addWidget(self.view_button_bar, 0, 4, 1, 4)
        layout3.addWidget(self.view_button_x, 1, 0, 1, 1)
        layout3.addWidget(self.view_button_xm, 1, 1, 1, 1)
        layout3.addWidget(label_a, 1, 4, 1, 1)
        layout3.addWidget(self.view_combo_a, 1, 5, 1, 3)
        layout3.addWidget(self.view_button_y, 2, 0, 1, 1)
        layout3.addWidget(self.view_button_ym, 2, 1, 1, 1)
        layout3.addWidget(label_b, 2, 4, 1, 1)
        layout3.addWidget(self.view_combo_b, 2, 5, 1, 3)
        layout3.addWidget(self.view_button_z, 3, 0, 1, 1)
        layout3.addWidget(self.view_button_zm, 3, 1, 1, 1)
        layout3.addWidget(label_c, 3, 4, 1, 1)
        layout3.addWidget(self.view_combo_c, 3, 5, 1, 3)

        panel4 = QWidget(parent)
        layout4 = Layout(panel4)
        layout4.setVerticalSpacing(2)
        layout4.addWidget(self.view_button_parallel, 1, 0, 1, 4)
        layout4.addWidget(self.view_button_grid, 1, 4, 1, 4)

        panel5 = QWidget(parent)
        layout5 = Layout(panel5)
        layout5.addWidget(label_axis, 0, 0, 1, 1)
        layout5.addWidget(self.view_combo_axis, 0, 1, 1, 3)
        layout5.addWidget(label_cell, 0, 4, 1, 1)
        layout5.addWidget(self.view_combo_cell, 0, 5, 1, 3)

        layout.addWidget(panel1, 0, 0, 1, 1)
        layout.addWidget(panel2, 1, 0, 1, 1)
        layout.addWidget(panel3, 2, 0, 1, 1)
        layout.addWidget(panel4, 3, 0, 1, 1)
        layout.addWidget(panel5, 4, 0, 1, 1)

        return panel

    # ==================================================
    def create_gui_dataset(self, parent):
        """
        Create dataset panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- dataset panel.

        :meta private:
        """
        panel = QWidget(parent)
        layout = Layout(panel)

        label_dataset = Label(parent, "DataSet", True)
        self.ds_button_edit = Button(parent, "edit")
        self.ds_button_clear = Button(parent, "clear")
        self.ds_button_load = Button(parent, "load")
        self.ds_button_save = Button(parent, "save")
        self.ds_button_screenshot = Button(parent, "screenshot")

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.addWidget(label_dataset, 0, 0, 1, 1)
        layout1.addWidget(self.ds_button_screenshot, 0, 1, 1, 1)
        layout1.addWidget(self.ds_button_edit, 1, 0, 1, 1)
        layout1.addWidget(self.ds_button_clear, 1, 1, 1, 1)
        layout1.addWidget(self.ds_button_load, 2, 0, 1, 1)
        layout1.addWidget(self.ds_button_save, 2, 1, 1, 1)

        layout.addWidget(panel1, 0, 0, 1, 1)

        return panel

    # ==================================================
    def create_gui_misc(self, parent):
        """
        Create misc. panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- misc. panel.

        :meta private:
        """
        panel = QWidget(parent)
        layout = Layout(panel)

        label_misc = Label(parent, "Misc", True)
        self.misc_button_info = Button(parent, "info")
        self.misc_button_pref = Button(parent, "preference")
        self.misc_button_about = Button(parent, "about")
        self.misc_button_log = Button(parent, "log")
        if check_multipie():
            self.misc_button_multipie = Button(parent, "MultiPie")

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.addWidget(label_misc, 0, 0, 1, 1)
        layout1.addWidget(self.misc_button_info, 0, 1, 1, 1)
        layout1.addWidget(self.misc_button_pref, 1, 0, 1, 1)
        layout1.addWidget(self.misc_button_about, 1, 1, 1, 1)
        layout1.addWidget(self.misc_button_log, 2, 0, 1, 1)
        if check_multipie():
            layout1.addWidget(self.misc_button_multipie, 2, 1, 1, 1)

        layout.addWidget(panel1, 0, 0, 1, 1)

        return panel

    # ==================================================
    def create_gui_debug(self, parent):
        """
        Create debug panel.

        Args:
            parent (QWidget): parent.

        Returns:
            - (QWidget) -- debug panel.

        :meta private:
        """
        panel = QWidget(parent)
        layout = Layout(panel)

        label_debug = Label(parent, "Debug", True)
        self.debug_button_camera = Button(parent, "camera")
        self.debug_button_actor = Button(parent, "actor")
        self.debug_button_data = Button(parent, "data")
        self.debug_button_status = Button(parent, "status")
        self.debug_button_preference = Button(parent, "pref")

        panel1 = QWidget(parent)
        layout1 = Layout(panel1)
        layout1.addWidget(label_debug, 0, 0, 1, 1)
        layout1.addWidget(self.debug_button_camera, 0, 1, 1, 1)
        layout1.addWidget(self.debug_button_data, 1, 0, 1, 1)
        layout1.addWidget(self.debug_button_actor, 1, 1, 1, 1)
        layout1.addWidget(self.debug_button_status, 2, 0, 1, 1)
        layout1.addWidget(self.debug_button_preference, 2, 1, 1, 1)

        layout.addWidget(panel1, 0, 0, 1, 1)

        return panel

    # ==================================================
    def _update_panel(self):
        """
        Update widget in panel.

        :meta private:
        """
        self._update_application()
        self._update_title()
        self._update_unit_cell()
        self._update_view()

    # ==================================================
    def _update_application(self):
        """
        Update application sytle.

        :meta private:
        """
        self.app.setStyle(self.pyvista_widget._preference["general"]["style"])
        font_type = self.pyvista_widget._preference["general"]["font"]
        size = self.pyvista_widget._preference["general"]["size"]
        font = QFont(font_type, size)
        self.app.setFont(font)

    # ==================================================
    def _update_title(self):
        """
        Update window title.

        :meta private:
        """
        title = self.pyvista_widget.window_title
        data_title = title.replace("QtDraw", "Dataset")
        self.setWindowTitle(title)
        self.pyvista_widget._tab_group_view.setWindowTitle(data_title)
        if self.multipie_dialog is not None:
            multipie_title = title.replace("QtDraw", "MultiPie Plugin")
            self.multipie_dialog.dialog.setWindowTitle(multipie_title)

    # ==================================================
    def _update_unit_cell(self):
        """
        Update unit cell widget.

        :meta private:
        """
        crystal = self.pyvista_widget._status["crystal"]

        self.uc_combo_crystal.setCurrentText(f"{crystal}")

        if crystal == "monoclinic":
            ro = [False, False, False, True, False, True]
        elif crystal == "orthorhombic":
            ro = [False, False, False, True, True, True]
        elif crystal in ["tetragonal", "trigonal", "hexagonal"]:
            ro = [False, True, False, True, True, True]
        elif crystal == "cubic":
            ro = [False, True, True, True, True, True]
        else:  # ""/"triclinic
            ro = [False, False, False, False, False, False]

        self.uc_edit_a.set_read_only(ro[0])
        self.uc_edit_b.set_read_only(ro[1])
        self.uc_edit_c.set_read_only(ro[2])
        self.uc_edit_alpha.set_read_only(ro[3])
        self.uc_edit_beta.set_read_only(ro[4])
        self.uc_edit_gamma.set_read_only(ro[5])

        self.uc_edit_origin.setText(f"{self.pyvista_widget.origin}")

        status_a = self.pyvista_widget._status["cell"]["a"]
        self.uc_edit_a.setText(str(status_a))
        status_b = self.pyvista_widget._status["cell"]["b"]
        self.uc_edit_b.setText(str(status_b))
        status_c = self.pyvista_widget._status["cell"]["c"]
        self.uc_edit_c.setText(str(status_c))
        status_alpha = self.pyvista_widget._status["cell"]["alpha"]
        self.uc_edit_alpha.setText(str(status_alpha))
        status_beta = self.pyvista_widget._status["cell"]["beta"]
        self.uc_edit_beta.setText(str(status_beta))
        status_gamma = self.pyvista_widget._status["cell"]["gamma"]
        self.uc_edit_gamma.setText(str(status_gamma))

        self.uc_label_volume.setText(f"Volume {self.pyvista_widget.cell_volume:.04f}")

    # ==================================================
    def _update_view(self):
        """
        Update view.

        :meta private:
        """
        self.view_button_clip.setChecked(self.pyvista_widget._status["clip"])

        self.view_button_repeat.setChecked(self.pyvista_widget._status["repeat"])

        self.view_edit_lower.setText(str(self.pyvista_widget._status["lower"]))
        self.view_edit_upper.setText(str(self.pyvista_widget._status["upper"]))

        self.view_combo_a.setCurrentText(str(self.pyvista_widget._status["view"][0]))
        self.view_combo_b.setCurrentText(str(self.pyvista_widget._status["view"][1]))
        self.view_combo_c.setCurrentText(str(self.pyvista_widget._status["view"][2]))

        self.view_button_bar.setChecked(self.pyvista_widget._status["bar"])
        self.view_button_parallel.setChecked(self.pyvista_widget._status["parallel_projection"])
        self.view_button_grid.setChecked(self.pyvista_widget._status["grid"])

        self.view_combo_axis.setCurrentText(self.pyvista_widget._status["axis_type"])
        self.view_combo_cell.setCurrentText(self.pyvista_widget._status["cell_mode"])

    # ==================================================
    def _set_crystal(self, crystal=None):
        """
        Set crystal.

        Args:
            crystal (str, optional): crystal.

        :meta private:
        """
        self.pyvista_widget.set_crystal(crystal)

        self._update_unit_cell()

    # ==================================================
    def _set_origin(self):
        """
        Set origin.

        :meta private:
        """
        origin = self.uc_edit_origin.text()
        if origin == "":
            origin = None
        self.pyvista_widget.set_origin(origin)

        self._update_unit_cell()

    # ==================================================
    def _set_a(self):
        """
        Set lattice constant, a.

        :meta private:
        """
        a = self.uc_edit_a.text()
        cell = None if a == "" else {"a": a}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_b(self):
        """
        Set lattice constant, b.

        :meta private:
        """
        b = self.uc_edit_b.text()
        cell = None if b == "" else {"b": b}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_c(self):
        """
        Set lattice constant, c.

        :meta private:
        """
        c = self.uc_edit_c.text()
        cell = None if c == "" else {"c": c}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_alpha(self):
        """
        Set lattice angle, alpha.

        :meta private:
        """
        alpha = self.uc_edit_alpha.text()
        cell = None if alpha == "" else {"alpha": alpha}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_beta(self):
        """
        Set lattice angle, beta.

        :meta private:
        """
        beta = self.uc_edit_beta.text()
        cell = None if beta == "" else {"beta": beta}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_gamma(self):
        """
        Set lattice angle, gamma.

        :meta private:
        """
        gamma = self.uc_edit_gamma.text()
        cell = None if gamma == "" else {"gamma": gamma}
        self.pyvista_widget.set_unit_cell(cell)

        self._update_unit_cell()

    # ==================================================
    def _set_clip(self, mode=None):
        """
        Set clip mode.

        Args:
            mode (bool): clip mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_clip(mode)

        self._update_view()

    # ==================================================
    def _set_repeat(self, mode=None):
        """
        Set repeat mode.

        Args:
            mode (bool): repeat mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_repeat(mode)

        self._update_view()

    # ==================================================
    def _set_bar(self, mode=None):
        """
        set bar.

        Args:
            mode (bool): scalar bar mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_bar(mode)

        self._update_view()

    # ==================================================
    def _set_lower(self):
        """
        Set lower bound.

        :meta private:
        """
        lower = self.view_edit_lower.text()
        if lower == "":
            lower = None
        self.pyvista_widget.set_range(lower, None)

        self._update_view()

    # ==================================================
    def _set_upper(self):
        """
        set upper bound.

        :meta private:
        """
        upper = self.view_edit_upper.text()
        if upper == "":
            upper = None
        self.pyvista_widget.set_range(None, upper)

        self._update_view()

    # ==================================================
    def _set_view_default(self):
        """
        set default.

        :meta private:
        """
        self.pyvista_widget.set_view()
        self._update_view()

    # ==================================================
    def _set_view(self, v, d):
        """
        set view direction.

        :meta private:
        """
        view = [
            int(self.view_combo_a.currentText()),
            int(self.view_combo_b.currentText()),
            int(self.view_combo_c.currentText()),
        ]
        view[d] = int(v)
        self.pyvista_widget.set_view(view)

        self._update_view()

    # ==================================================
    def _set_parallel(self, mode=None):
        """
        set parallel projection mode.

        Args:
            mode (bool): parallel projection mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_parallel_projection(mode)

        self._update_view()

    # ==================================================
    def _set_grid(self, mode=None):
        """
        set grid mode.

        Args:
            mode (bool): grid mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_grid(mode)

        self._update_view()

    # ==================================================
    def _set_axis_type(self, mode=None):
        """
        set axis type.

        Args:
            mode (str): axis type.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_axis(mode)

        self._update_view()

    # ==================================================
    def _set_cell_mode(self, mode=None):
        """
        set cell mode.

        Args:
            mode (str): cell mode.

        Note:
            - if mode is None, default is used.

        :meta private:
        """
        self.pyvista_widget.set_cell(mode)

        self._update_view()

    # ==================================================
    def create_connection(self):
        """
        Create connections.

        :meta private:
        """
        # unit cell panel.
        self.uc_combo_crystal.currentTextChanged.connect(self._set_crystal)
        self.uc_edit_origin.returnPressed.connect(self._set_origin)
        self.uc_edit_a.returnPressed.connect(self._set_a)
        self.uc_edit_b.returnPressed.connect(self._set_b)
        self.uc_edit_c.returnPressed.connect(self._set_c)
        self.uc_edit_alpha.returnPressed.connect(self._set_alpha)
        self.uc_edit_beta.returnPressed.connect(self._set_beta)
        self.uc_edit_gamma.returnPressed.connect(self._set_gamma)

        # view panel.
        self.view_button_clip.toggled.connect(self._set_clip)
        self.view_button_repeat.toggled.connect(self._set_repeat)
        self.view_button_nonrepeat.pressed.connect(self._nonrepeat)
        self.view_edit_lower.returnPressed.connect(self._set_lower)
        self.view_edit_upper.returnPressed.connect(self._set_upper)
        self.view_button_bar.toggled.connect(self._set_bar)
        self.view_button_default.pressed.connect(self._set_view_default)
        self.view_button_x.pressed.connect(self.pyvista_widget.view_yz)
        self.view_button_y.pressed.connect(self.pyvista_widget.view_zx)
        self.view_button_z.pressed.connect(self.pyvista_widget.view_xy)
        self.view_button_xm.pressed.connect(self.pyvista_widget.view_zy)
        self.view_button_ym.pressed.connect(self.pyvista_widget.view_xz)
        self.view_button_zm.pressed.connect(self.pyvista_widget.view_yx)
        self.view_combo_a.currentTextChanged.connect(lambda x: self._set_view(x, 0))
        self.view_combo_b.currentTextChanged.connect(lambda x: self._set_view(x, 1))
        self.view_combo_c.currentTextChanged.connect(lambda x: self._set_view(x, 2))
        self.view_button_parallel.toggled.connect(self._set_parallel)
        self.view_button_grid.toggled.connect(self._set_grid)
        self.view_combo_axis.currentTextChanged.connect(self._set_axis_type)
        self.view_combo_cell.currentTextChanged.connect(self._set_cell_mode)

        # dataset panel.
        self.ds_button_edit.pressed.connect(self.pyvista_widget.open_tab_group_view)
        self.ds_button_clear.pressed.connect(self._clear_data)
        self.ds_button_load.pressed.connect(self.open_file)
        self.ds_button_save.pressed.connect(self.save_file)
        self.ds_button_screenshot.pressed.connect(self._save_screenshot)

        # misc panel.
        self.misc_button_info.pressed.connect(lambda: self.info_dialog.show())
        self.misc_button_pref.pressed.connect(self._show_preference)
        self.misc_button_about.pressed.connect(self._show_about)
        self.misc_button_log.pressed.connect(self.logger.show)
        if check_multipie():
            self.misc_button_multipie.pressed.connect(self._show_multipie)

        # debug panel.
        if self.debug:
            self.debug_button_camera.pressed.connect(self._show_camera_info)
            self.debug_button_data.pressed.connect(self._show_raw_data)
            self.debug_button_actor.pressed.connect(self._show_actor_list)
            self.debug_button_status.pressed.connect(self._show_status_data)
            self.debug_button_preference.pressed.connect(self._show_preference_data)

        # info message.
        self.pyvista_widget.message.connect(self.write_info)

    # ==================================================
    def _nonrepeat(self):
        """
        Transform data into non-repeat data.

        :meta private:
        """
        self.pyvista_widget.nonrepeat_data()

    # ==================================================
    def _show_preference(self):
        """
        Show preference panel.

        :meta private:
        """
        self.pyvista_widget.save_current()
        self.pref_dialog = PreferenceDialog(self.pyvista_widget, self)
        status = self.pref_dialog.exec()  # open as modal mode.
        self.sender().setDown(False)  # reset push button.
        if status == QDialog.Accepted:
            self.app.setStyleSheet(create_style_sheet(self.pyvista_widget._preference["general"]["size"]))
            self.pyvista_widget.refresh()
            self.pyvista_widget.redraw()
            self._update_panel()
        else:
            self.pyvista_widget.restore()
            self.app.setStyleSheet(create_style_sheet(self.pyvista_widget._preference["general"]["size"]))
            self._update_panel()

    # ==================================================
    def _show_about(self):
        """
        Show about panel.

        :meta private:
        """
        self.about_dialog = AboutDialog(self.pyvista_widget, self)
        self.about_dialog.exec()  # open as modal mode.
        self.sender().setDown(False)  # reset push button.

    # ==================================================
    def _show_multipie(self):
        """
        Show MultiPie panel.

        :meta private:
        """
        if check_multipie():
            from qtdraw.multipie.plugin_multipie import MultiPiePlugin

            if self.multipie_dialog is None:
                self.multipie_dialog = MultiPiePlugin(self)
            else:
                self.multipie_dialog.dialog.show()
            self.sender().setDown(False)  # reset push button.

    # ==================================================
    def _show_status_data(self):
        """
        Show status data dialog.

        :meta private:
        """
        s = ""
        for key, val in self.pyvista_widget._status.items():
            if key != "plus" and key != "multipie":
                s += f"{key}: {val}\n"

        s += "\n=== multipie ===\n"
        if "version" in self.pyvista_widget._status["multipie"].keys():
            ver = self.pyvista_widget._status["multipie"]["version"]
            s += f"version: {ver}\n"
        for key, val in self.pyvista_widget._status["multipie"].items():
            if key != "plus":
                if key != "version":
                    s += f"--- {key} ---\n"
                    for k, v in val.items():
                        s += f"{k}: {v}\n"

        s += "\n=== plus ===\n"
        for key, val in self.pyvista_widget._status["plus"].items():
            s += f"{key}: {val}\n"

        s += "\n=== multipie.plus ===\n"
        for key, val in self.pyvista_widget._status["multipie"]["plus"].items():
            s += f"{key}: {val}\n"
        s = s[:-1]

        self.status_dialog = LogWidget("Status Data", None)
        self.status_dialog.set_text(s)
        self.status_dialog.show()

    # ==================================================
    def _show_preference_data(self):
        """
        Show preference data dialog.

        :meta private:
        """
        s = ""
        for key, val in self.pyvista_widget._preference.items():
            s += f"=== {key} ===\n"
            for name, v in val.items():
                s += f"{name}: {v}\n"
        s = s[:-1]
        self.pref_data_dialog = LogWidget("Preference Data", None)
        self.pref_data_dialog.set_text(s)
        self.pref_data_dialog.show()

    # ==================================================
    def _show_actor_list(self):
        """
        Show actor list dialog.

        :meta private:
        """
        s = ""
        for i in self.pyvista_widget.actor_list:
            s += i + "\n"
        s = s[:-1]
        self.actor_dialog = LogWidget("Actor Data", None)
        self.actor_dialog.set_text(s)
        self.actor_dialog.show()

    # ==================================================
    def _show_raw_data(self):
        """
        Show raw data dialog.

        :meta private:
        """
        s = ""
        for object_type, model in self.pyvista_widget._data.items():
            s += f"=== {object_type} ===\n"
            data = model.tolist()
            for row in data:
                s += str(row) + "\n"
        s = s[:-1]
        self.data_dialog = LogWidget("Raw Data", None)
        self.data_dialog.set_text(s)
        self.data_dialog.show()

    # ==================================================
    def _show_camera_info(self):
        """
        Show camera info. dialog.

        :meta private:
        """
        camera = self.pyvista_widget.get_camera_info()
        s = ""
        for key, val in camera.items():
            s += f"{key}: {val}\n"
        s = s[:-1]
        self.camera_dialog = LogWidget("Camera Info", None)
        self.camera_dialog.set_text(s)
        self.camera_dialog.show()

    # ==================================================
    def _clear_data(self):
        """
        Clear data (actor and data).

        :meta private:
        """
        ret = QMessageBox.question(self, "", "Are you sure ?", QMessageBox.Cancel, QMessageBox.Ok)
        self.sender().setDown(False)  # reset push button.
        if ret == QMessageBox.Ok:
            self.clear_data()

    # ==================================================
    def clear_data(self):
        """
        Clear data (actor and data).
        """
        self.pyvista_widget.reload()
        self._update_panel()
        if self.multipie_dialog is not None:
            self.multipie_dialog.clear_counter()

    # ==================================================
    def exec(self):
        """
        Execute QtDraw.

        :meta private:
        """
        self.app.exec()

    # ==================================================
    def closeEvent(self, event):
        """
        Close with dialog.

        Args:
            event (Event): event.

        :meta private:
        """
        ret = QMessageBox.question(self, "", "Quit QtDraw ?", QMessageBox.Cancel, QMessageBox.Ok)
        if ret != QMessageBox.Ok:
            event.ignore()
        else:
            self.close()

    # ==================================================
    def close(self):
        """
        Close dialogs.

        :meta private:
        """
        if self.debug:
            if self.actor_dialog is not None:
                self.actor_dialog.close()
            if self.data_dialog is not None:
                self.data_dialog.close()
            if self.status_dialog is not None:
                self.status_dialog.close()
            if self.pref_data_dialog is not None:
                self.pref_data_dialog.close()
            if self.camera_dialog is not None:
                self.camera_dialog.close()
        if self.multipie_dialog is not None:
            self.multipie_dialog.dialog.close()
        self.logger.close()
        self.info_dialog.close()
        self.pyvista_widget.close()
        super().close()

    # ==================================================
    def update_status(self, key, value):
        """
        Update status.

        Args:
            key (str): status key.
            value (Any): value.

        :meta private:
        """
        self.pyvista_widget.update_status(key, value)
        self._update_panel()

    # ==================================================
    def update_preference(self, category, key, value):
        """
        Update preference.

        Args:
            category (str): category.
            key (str): status key.
            value (Any): value.

        :meta private:
        """
        self.pyvista_widget.update_preference(category, key, value)
        self._update_panel()

    # ==================================================
    def write_info(self, text):
        """
        Write text message into Info dialog.

        Args:
            text (str): message.

        :meta private:
        """
        self.info_dialog.append_text(text)

    # ==================================================
    def add_site(self, size=None, color=None, opacity=None, position=None, cell=None, name=None, label=None, margin=None):
        """
        Add site.

        Args:
            size (float, optional): site size. (default: 0.1)
            color (str, optional): site color. (default: darkseagreen)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_site(size, color, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_bond(
        self,
        direction=None,
        width=None,
        color=None,
        color2=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add bond.

        Args:
            direction (str, optional): bond direction. (default: [0,0,1])
            width (float, optional): bond width. (default: 0.02)
            color (str, optional): bond color (tail side). (default: silver)
            color2 (str, optional): bond color (head side). (default: silver)
            cartesian (bool, optional): cartesian coordinate for direction ? (default: False)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
            - bond position is at center.
        """
        self.pyvista_widget.add_bond(direction, width, color, color2, cartesian, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_vector(
        self,
        direction=None,
        length=None,
        width=None,
        offset=None,
        color=None,
        cartesian=None,
        shaft_R=None,
        tip_R=None,
        tip_length=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add vector.

        Args:
            direction (str, optional): vector direction. (default: [0,0,1])
            length (float, optional): vector length. (default: 1.0)
            width (float, optional): vector width. (default: 0.02)
            offset (float, optional): vector offset. (default: -0.43)
            color (str, optional): vector color. (default: orange)
            cartesian (bool, optional): cartesian coordinate for direction ? (default: True)
            shaft_R (float, optional): shaft radius. (default: 1.0)
            tip_R (float, optional): tip radius. (default: 2.0)
            tip_length (float, optional): tip length. (default: 0.25)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
            - if length is negative, norm of direction multiplied by |length| is used.
        """
        self.pyvista_widget.add_vector(
            direction,
            length,
            width,
            offset,
            color,
            cartesian,
            shaft_R,
            tip_R,
            tip_length,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_orbital(
        self,
        shape=None,
        surface=None,
        size=None,
        range=None,
        color=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add orbital.

        Args:
            shape (str, optional): orbital shape polynomial in terms of (x,y,z,r). (default: 3z**2-r**2)
            surface (str, optional): orbital colormap polynomial in terms of (x,y,z,r). (default: "")
            size (float, optional): orbital size. (default: 0.5)
            range (list, optional): plot range of [th,phi]. (default: [[0,180],[0,360]])
            color (str, optional): orbital color or colormap. (default: Wistia)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
            - if surface is "", the same one of shape is used.
            - if size is positive, max. value is equivalent to size.
            - if size is negative, abs. value is scaled by size.
        """
        self.pyvista_widget.add_orbital(shape, surface, size, range, color, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_stream(
        self,
        shape=None,
        vector=None,
        size=None,
        range=None,
        division=None,
        length=None,
        width=None,
        offset=None,
        abs_scale=None,
        color=None,
        component=None,
        shaft_R=None,
        tip_R=None,
        tip_length=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add stream.

        Args:
            shape (str, optional): shape polynomial on which stream put. (default: 1)
            vector (str, optional): stream polynomials for [x,y,z] components. (default: [x,y,z])
            size (float, optional): shape size. (default: 0.5)
            range (list, optional): plot range of [th,phi]. (default: [[0,180],[0,360]])
            division (list, optional): division of stream of [th,phi]. (default: [9,18])
            length (float, optional): stream arrow size. (default: 0.1)
            width (float, optional): stream arrow width. (default: 0.01)
            offset (float, optional): stream arrow offset. (default: -0.43)
            abs_scale (bool, optional): stream arrow scaled ? (default: False)
            color (str, optional): stream arrow color or colormap. (default: coolwarm)
            component (str, optional): use component or abs, "x/y/z/abs". (default: abs)
            shaft_R (float, optional): shaft radius. (default: 1.0)
            tip_R (float, optional): tip radius. (default: 2.0)
            tip_length (float, optional): tip length. (default: 0.25)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
            - if size is negative, shape is normalized.
        """
        self.pyvista_widget.add_stream(
            shape,
            vector,
            size,
            range,
            division,
            length,
            width,
            offset,
            abs_scale,
            color,
            component,
            shaft_R,
            tip_R,
            tip_length,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_line(
        self,
        direction=None,
        width=None,
        arrow1=None,
        arrow2=None,
        tip_R=None,
        tip_length=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add line.

        Args:
            direction (str, optional): line direction. (default: [0,0,1])
            width (float, optional): line width. (default: 0.02)
            arrow1 (bool, optional): arrow (tail side) ? (default: False)
            arrow2 (bool, optional): arrow (head side) ? (default: False)
            tip_R (float, optional): arrow tip radius. (default: 2.0)
            tip_length (float, optional): arrow tip length. (default: 0.1)
            color (str, optional): line color. (default: strawberry)
            cartesian (bool, optional): cartesian coordinate for direction ? (default: False)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_line(
            direction, width, arrow1, arrow2, tip_R, tip_length, color, cartesian, opacity, position, cell, name, label, margin
        )

    # ==================================================
    def add_plane(
        self,
        normal=None,
        x_size=None,
        y_size=None,
        color=None,
        width=None,
        grid=None,
        grid_color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add plane.

        Args:
            normal (str, optional): plane normal vector. (default: [0,0,1])
            x_size (float, optional): plane x size. (default: 1.0)
            y_size (float, optional): plane y size. (default: 1.0)
            color (str, optional): plane color. (default: sky)
            width (float, optional): plane grid width. (default: 2.0)
            grid (bool, optional): grid ? (default: False)
            grid_color (str, optional): grid color. (default: black)
            cartesian (bool, optional): cartesian coordinate for normal ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_plane(
            normal, x_size, y_size, color, width, grid, grid_color, cartesian, opacity, position, cell, name, label, margin
        )

    # ==================================================
    def add_circle(
        self,
        normal=None,
        size=None,
        color=None,
        width=None,
        edge=None,
        edge_color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add circle.

        Args:
            normal (str, optional): circle normal vector. (default: [0,0,1])
            size (float, optional): circle size. (default: 0.5)
            color (str, optional): circle color. (default: salmon)
            width (float, optional): edge width. (default: 2.0)
            edge (bool, optional): edge ? (default: True)
            edge_color (str, optional): edge color. (default: black)
            cartesian (bool, optional): cartesian coordinate for normal ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_circle(
            normal, size, color, width, edge, edge_color, cartesian, opacity, position, cell, name, label, margin
        )

    # ==================================================
    def add_torus(
        self,
        normal=None,
        size=None,
        width=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add torus.

        Args:
            normal (str, optional): torus normal vector. (default: [0,0,1])
            size (float, optional): torus size. (default: 0.5)
            width (float, optional): torus width. (default: 0.15)
            color (str, optional): torus color. (default: cantaloupe)
            cartesian (bool, optional): cartesian coordinate for normal ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_torus(normal, size, width, color, cartesian, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_ellipsoid(
        self,
        normal=None,
        x_size=None,
        y_size=None,
        z_size=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add ellipsoid.

        Args:
            normal (str, optional): ellipsoid normal vector. (default: [0,0,1])
            x_size (float, optional): ellipsoid x size. (default: 0.5)
            y_size (float, optional): ellipsoid y size. (default: 0.4)
            z_size (float, optional): ellipsoid z size. (default: 0.3)
            color (str, optional): ellipsoid color. (default: cornflowerblue)
            cartesian (bool, optional): cartesian coordinate for normal ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_ellipsoid(
            normal, x_size, y_size, z_size, color, cartesian, opacity, position, cell, name, label, margin
        )

    # ==================================================
    def add_toroid(
        self,
        normal=None,
        size=None,
        width=None,
        x_scale=None,
        y_scale=None,
        z_scale=None,
        ring_shape=None,
        tube_shape=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add toroid.

        Args:
            normal (str, optional): toroid normal vector. (default: [0,0,1])
            size (float, optional): toroid size. (default: 0.5)
            width (float, optional): toroid width. (default: 0.15)
            x_scale (float, optional): toroid x scale. (default: 1.0)
            y_scale (float, optional): toroid y scale. (default: 1.0)
            z_scale (float, optional): toroid z scale. (default: 1.0)
            ring_shape (float, optional): toroid ring shape. (default: 0.3)
            tube_shape (float, optional): toroid tube shape. (default: 0.3)
            color (str, optional): toroid color. (default: tan)
            cartesian (bool, optional): cartesian coordinate for normal ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_toroid(
            normal,
            size,
            width,
            x_scale,
            y_scale,
            z_scale,
            ring_shape,
            tube_shape,
            color,
            cartesian,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_box(
        self,
        a1=None,
        a2=None,
        a3=None,
        width=None,
        edge=None,
        edge_color=None,
        wireframe=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add box.

        Args:
            a1 (str, optional): box a1 vector. (default: [1,0,0])
            a2 (str, optional): box a2 vector. (default: [0,1,0])
            a3 (str, optional): box a3 vector. (default: [0,0,1])
            width (float, optional): edge width. (default: 2.0)
            edge (bool, optional): edge ? (default: True)
            edge_color (str, optional): edge color. (default: black)
            wireframe (bool, optional): wireframe ? (default: False)
            color (str, optional): box color. (default: yellowgreen)
            cartesian (bool, optional): cartesian coordinate for shape ? (default: False)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_box(
            a1, a2, a3, width, edge, edge_color, wireframe, color, cartesian, opacity, position, cell, name, label, margin
        )

    # ==================================================
    def add_polygon(
        self,
        point=None,
        connectivity=None,
        width=None,
        edge=None,
        edge_color=None,
        wireframe=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add polygon.

        Args:
            point (str, optional): polygon vertex points. (default: [[0,0,0],[0.8,0,0],[0,0.6,0],[0,0,0.4],[0.6,0.6,0]])
            connectivity (str, optional): connectivity. (default: [[0,1,4,2],[0,1,3],[1,4,3],[2,0,3],[2,3,4]])
            width (float, optional): edge width. (default: 2.0)
            edge (bool, optional): edge ? (default: True)
            edge_color (str, optional): edge color. (default: black)
            wireframe (bool, optional): wireframe ? (default: False)
            color (str, optional): polygon color. (default: aluminum)
            cartesian (bool, optional): cartesian coordinate for shape ? (default: False)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_polygon(
            point,
            connectivity,
            width,
            edge,
            edge_color,
            wireframe,
            color,
            cartesian,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_spline(
        self,
        point=None,
        width=None,
        n_interp=None,
        closed=None,
        natural=None,
        arrow1=None,
        arrow2=None,
        tip_R=None,
        tip_length=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add spline.

        Args:
            point (str, optional): spline vertex points. (default: [[0,0,0],[1,0,1],[0,1,2]])
            width (float, optional): spline width. (default: 0.01)
            n_interp (int, optional): spline interpolation points. (default: 500)
            closed (bool, optional): closed curve ? (default: False)
            natural (bool, optional): natural edge ? (default: True)
            arrow1 (bool, optional): arrow (tail side) ? (default: False)
            arrow2 (bool, optional): arrow (head side) ? (default: False)
            tip_R (float, optional): arrow tip radius. (default: 2.0)
            tip_length (float, optional): arrow tip length. (default: 0.1)
            color (str, optional): spline color. (default: banana)
            cartesian (bool, optional): cartesian coordinate for point ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_spline(
            point,
            width,
            n_interp,
            closed,
            natural,
            arrow1,
            arrow2,
            tip_R,
            tip_length,
            color,
            cartesian,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_spline_t(
        self,
        point=None,
        t_range=None,
        width=None,
        n_interp=None,
        closed=None,
        natural=None,
        arrow1=None,
        arrow2=None,
        tip_R=None,
        tip_length=None,
        color=None,
        cartesian=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add spline (parametric).

        Args:
            point (str, optional): spline function vector in terms of t. (default: [cos(2 pi t), sin(2 pi t), t/2])
            t_range (str, optional): spline t range [t0,t1,dt]. (default: [0,1,0.05])
            width (float, optional): spline width. (default: 0.01)
            n_interp (int, optional): spline interpolation points. (default: 500)
            closed (bool, optional): closed curve ? (default: False)
            natural (bool, optional): natural edge ? (default: True)
            arrow1 (bool, optional): arrow (tail side) ? (default: False)
            arrow2 (bool, optional): arrow (head side) ? (default: False)
            tip_R (float, optional): arrow tip radius. (default: 2.0)
            tip_length (float, optional): arrow tip length. (default: 0.1)
            color (str, optional): splline color. (default: crimson)
            cartesian (bool, optional): cartesian coordinate for point ? (default: True)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_spline_t(
            point,
            t_range,
            width,
            n_interp,
            closed,
            natural,
            arrow1,
            arrow2,
            tip_R,
            tip_length,
            color,
            cartesian,
            opacity,
            position,
            cell,
            name,
            label,
            margin,
        )

    # ==================================================
    def add_text3d(
        self,
        text=None,
        size=None,
        view=None,
        depth=None,
        offset=None,
        color=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add text 3d.

        Args:
            text (str, optional): text. (default: text)
            size (float, optional): text size. (default: 0.3)
            view (str, optional): text viewpoint. (default: [0,0,1])
            depth (float, optional): text depth. (default: 0.2)
            offset (str, optional): text offset. (default: [0,0,0])
            color (str, optional): text color. (default: iron)
            opacity (float, optional): opacity, [0,1]. (default: 1.0)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_text3d(text, size, view, depth, offset, color, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_isosurface(
        self,
        data=None,
        value=None,
        surface=None,
        color=None,
        color_range=None,
        opacity=None,
        position=None,
        cell=None,
        name=None,
        label=None,
        margin=None,
    ):
        """
        Add isosurface.

        Args:
            data (str, optional): data (file) name for grid data. (default: "")
            value (str, optional): isosurface values. (default: [0.5])
            surface (str, optional): surface value name. (default: "")
            color (str, optional): text color. (default: white)
            color_range (str, optional): color range. (default: [0,1])
            opacity (float, optional): opacity, [0,1]. (default: 0.8)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            cell (str, optional): cell, [nx,ny,nz]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            label (str, optional): label. (default: label)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
            - if filename is tuple, (name, dict), use dict data as name.
        """
        self.pyvista_widget.add_isosurface(data, value, surface, color, color_range, opacity, position, cell, name, label, margin)

    # ==================================================
    def add_caption(self, caption=None, size=None, bold=None, color=None, position=None, cell=None, name=None, margin=None):
        """
        Add caption.

        Args:
            caption (str, optional): caption list. (default: [A,B,C])
            size (int, optional): caption size. (default: 18)
            bold (bool, optional): bold ? (default: True)
            color (str, optional): caption color. (default: licorice)
            position (str, optional): position of each caption. (default: [[0,0,0],[1,0,0],[1,1,0]])
            cell (str, optional): cell. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)
            margin (int, optional): label margin. (default: 3)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_caption(caption, size, bold, color, position, cell, name, margin)

    # ==================================================
    def add_text2d(self, caption=None, size=None, color=None, font=None, position=None, name=None):
        """
        Add text 2d.

        Args:
            caption (str, optional): caption. (default: text)
            size (int, optional): caption size. (default: 8)
            color (str, optional): caption color. (default: licorice)
            font (str, optional): caption font. (default: arial)
            position (str, optional): position in cell, [x,y,z]. (default: [0,0,0])
            name (str, optional): name of group. (default: untitled)

        Note:
            - if keyword is None, default value is used.
        """
        self.pyvista_widget.add_text2d(caption, size, color, font, position, name)

    # ==================================================
    def plot_orbital_from_data(
        self,
        name,
        shape,
        surface=None,
        size=1.0,
        point_size=0.03,
        spherical_plot=False,
        color="coolwarm",
        opacity=1.0,
        position=None,
    ):
        """
        Plot orbital from data.

        Args:
            name (str): plot name.
            shape (ndarray): (x,y,z) orbital shape.
            surface (ndarray, optional): (x,y,z) orbital colormap.
            size (float, optional): orbital size.
            point_size(float, optional): point size.
            spherical_plot (bool, optional): spherical-like plot ?
            color (str, optional): orbital color or colormap.
            opacity (float, optional): opacity, [0,1].
            position (ndarray): position (transformed).

        Note:
            - if surface is None, the same one of shape is used.
            - if size is positive, max. value is equivalent to size.
            - if size is negative, abs. value is scaled by size.
            - if point_size is None, no point is shown.
            - to remove object, use remove_actor(name).
        """
        self.pyvista_widget.plot_orbital_from_data(
            name, shape, surface, size, point_size, spherical_plot, color, opacity, position
        )

    # ==================================================
    def plot_stream_from_data(
        self,
        name,
        vector,
        shape,
        surface=None,
        size=1.0,
        length=0.2,
        width=0.01,
        offset=-0.43,
        abs_scale=False,
        color="coolwarm",
        component="abs",
        spherical_plot=False,
        shaft_R=1.0,
        tip_R=2.0,
        tip_length=0.25,
        opacity=1.0,
        position=None,
    ):
        """
        Plot stream from data (vectors in cartesian coordinate).

        Args:
            name (str): object name.
            vector (ndarray): stream vector [vx(x,y,z),vy(x,y,z),vz(x,y,z)] (cartesina).
            shape (ndarray): f(x,y,z) shape on which stream put (cartesian).
            surface (ndarray, optional): (x,y,z) orbital colormap.
            size (float, optional): shape size.
            length (float, optional): stream arrow size.
            width (float, optional): stream arrow width.
            offset (float, optional): stream arrow offest.
            abs_scale (bool, optional): stream arrow scaled ?
            color (str, optional): stream arrow color or colormap.
            component (str, optional): use component or abs, "x/y/z/abs".
            spherical_plot (bool, optional): spherical-like plot ?
            shaft_R (float, optional) :shaft radius.
            tip_R (float, optional): tip radius.
            tip_length (float, optional): tip length.
            opacity (float, optional): opacity, [0,1].
            position (ndarray): position (transformed).

        Note:
            - if size is negative, shape is normalized.
            - to remove object, use remove_actor(name).
        """
        self.pyvista_widget.plot_stream_from_data(
            name,
            vector,
            shape,
            surface,
            size,
            length,
            width,
            offset,
            abs_scale,
            color,
            component,
            spherical_plot,
            shaft_R,
            tip_R,
            tip_length,
            opacity,
            position,
        )

    # ==================================================
    def set_model(self, name=None):
        """
        Set model name.

        Args:
            name (str, optional): model name.

        Note:
            - if name is None, default is used.
        """
        self.pyvista_widget.set_model(name)
        self._update_title()

    # ==================================================
    def set_unit_cell(self, cell=None):
        """
        Set unit cell.

        Args:
            cell (dict, optional): unit cell, {a, b, c, alpha, beta, gamma}.

        Note:
            - if cell is None, default is used.
            - set cell.
        """
        view = self.pyvista_widget._status["view"]
        self.pyvista_widget.set_unit_cell(cell)
        self._update_unit_cell()
        self._update_view()
        self.set_view(view)

    # ==================================================
    def set_crystal(self, crystal=None):
        """
        Set crystal.

        Args:
            crystal (str, optional): crystal, "triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic".

        Note:
            - if crystal is None, default is used.
            - set unit cell.
        """
        view = self.pyvista_widget._status["view"]
        self.pyvista_widget.set_crystal(crystal)
        self._update_view()
        self.set_view(view)

    # ==================================================
    def set_origin(self, origin=None):
        """
        Set origin.

        Args:
            origin (list or str, optional): origin.

        Note:
            - if origin is None, default is used.
            - set cell.
        """
        view = self.pyvista_widget._status["view"]
        self.pyvista_widget.set_origin(origin)
        self._update_view()
        self.set_view(view)

    # ==================================================
    def set_clip(self, mode=None):
        """
        Set clip mode.

        Args:
            mode (bool, optional): clip object ?

        Note:
            - if mode is None, default is used.
        """
        self._set_clip(mode)

    # ==================================================
    def set_repeat(self, mode=None):
        """
        Set repeat mode.

        Args:
            mode (bool, optional): repeat mode.

        Note:
            - if mode is None, default is used.
            - repeat data.
        """
        self._set_repeat(mode)

    # ==================================================
    def set_nonrepeat(self):
        """
        Transform data to non-repeat data.
        """
        self.pyvista_widget.nonrepeat_data()

    # ==================================================
    def set_range(self, lower=None, upper=None):
        """
        Set cell range.

        Args:
            lower (list or str, optional): lower bound, [float].
            upper (list or str, optional): upper bound, [float].

        Note:
            - if lower/upper is None, default is used.
            - set cell.
        """
        view = self.pyvista_widget._status["view"]
        self.pyvista_widget.set_range(lower, upper)
        self._update_view()
        self.set_view(view)

    # ==================================================
    def set_view(self, view=None):
        """
        Set view point.

        Args:
            view (list or str, optional): view indices of [a1,a2,a3], [int].

        Note:
            - if view is None, default is used.
        """
        if view is None:
            self._set_view_default()
            return

        if type(view) == str:
            view = view.strip("[]").split(",")
        self.view_combo_a.setCurrentText(str(view[0]))
        self.view_combo_b.setCurrentText(str(view[1]))
        self.view_combo_c.setCurrentText(str(view[2]))

    # ==================================================
    def set_parallel_projection(self, mode=None):
        """
        Set parallel projection mode.

        Args:
            mode (bool, optional): use parallel projection ?

        Note:
            - if mode is None, default is used.
        """
        view = self.pyvista_widget._status["view"]
        self._set_parallel(mode)
        self.set_view(view)

    # ==================================================
    def set_grid(self, mode=None):
        """
        Set grid mode.

        Args:
            mode (bool, optional): use grid ?

        Note:
            - if mode is None, default is used.
        """
        self._set_grid(mode)

    # ==================================================
    def set_bar(self, mode=None):
        """
        Set scalar bar mode.

        Args:
            mode (bool, optional): show scalar bar ?

        Note:
            - if mode is None, default is used.
        """
        self._set_bar(mode)

    # ==================================================
    def set_axis(self, axis_type=None):
        """
        Set axis widget.

        Args:
            axis_type (str, optional): axis type, "on/axis/full/off".

        Note:
            - if axis_type is None, default is used.
        """
        self._set_axis_type(axis_type)

    # ==================================================
    def set_cell(self, mode=None):
        """
        Set unit cell.

        Args:
            mode (str, optional): mode "single/all/off.

        Note:
            - if mode is None, default is used.
        """
        self._set_cell_mode(mode)

    # ==================================================
    def add_mesh(
        self,
        mesh,
        color=None,
        style=None,
        scalars=None,
        clim=None,
        show_edges=None,
        edge_color=None,
        point_size=None,
        line_width=None,
        opacity=None,
        flip_scalars=False,
        lighting=None,
        n_colors=256,
        interpolate_before_map=None,
        cmap=None,
        label=None,
        reset_camera=None,
        scalar_bar_args=None,
        show_scalar_bar=None,
        multi_colors=False,
        name=None,
        texture=None,
        render_points_as_spheres=None,
        render_lines_as_tubes=None,
        smooth_shading=None,
        split_sharp_edges=None,
        ambient=None,
        diffuse=None,
        specular=None,
        specular_power=None,
        nan_color=None,
        nan_opacity=1.0,
        culling=None,
        rgb=None,
        categories=False,
        silhouette=None,
        use_transparency=False,
        below_color=None,
        above_color=None,
        annotations=None,
        pickable=True,
        preference="point",
        log_scale=False,
        pbr=None,
        metallic=None,
        roughness=None,
        render=True,
        user_matrix=None,
        component=None,
        emissive=None,
        copy_mesh=False,
        backface_params=None,
        show_vertices=None,
        edge_opacity=None,
        **kwargs,
    ):
        """
        Add any PyVista/VTK mesh or dataset that PyVista can wrap to the scene.

        See PyVista manual.
        """
        self.pyvista_widget.add_mesh(
            mesh,
            color,
            style,
            scalars,
            clim,
            show_edges,
            edge_color,
            point_size,
            line_width,
            opacity,
            flip_scalars,
            lighting,
            n_colors,
            interpolate_before_map,
            cmap,
            label,
            reset_camera,
            scalar_bar_args,
            show_scalar_bar,
            multi_colors,
            name,
            texture,
            render_points_as_spheres,
            render_lines_as_tubes,
            smooth_shading,
            split_sharp_edges,
            ambient,
            diffuse,
            specular,
            specular_power,
            nan_color,
            nan_opacity,
            culling,
            rgb,
            categories,
            silhouette,
            use_transparency,
            below_color,
            above_color,
            annotations,
            pickable,
            preference,
            log_scale,
            pbr,
            metallic,
            roughness,
            render,
            user_matrix,
            component,
            emissive,
            copy_mesh,
            backface_params,
            show_vertices,
            edge_opacity,
            **kwargs,
        )

    # ==================================================
    def remove_actor(self, actor, reset_camera=False, render=True):
        """
        Remove actor.

        Args:
            actor (str, vtk.vtkActor, list or tuple): actor name.
            reset_camera (bool, optional): reset camera to show all actors ?
            render (bool, optional): render upon actor removal.

        Returns:
            - (bool) -- True if actor is removed.
        """
        return self.pyvista_widget.remove_actor(actor, reset_camera, render)

    # ==================================================
    def load(self, filename):
        """
        Load all info.

        Args:
            filename (str): full file name.
        """
        self.pyvista_widget.load(filename)

    # ==================================================
    def save(self, filename):
        """
        save all info.

        Args:
            filename (str): full file name.
        """
        self.pyvista_widget.save(filename)

    # ==================================================
    # MultiPie interface
    # ==================================================
    def mp_set_group(self, tag=None):
        """
        MultiPie: Set point/sapce group.

        Args:
            tag (str, optional): group tag in Schoenflies notation [default: C1].
        """
        if not check_multipie():
            raise Exception("MultiPie is not installed.")

        if self.multipie_dialog is not None:
            self.multipie_dialog.dialog.close()
            del self.multipie_dialog
            self.multipie_dialog = None

        if tag is None:
            tag = "C1"

        multipie = {"group": {"group": tag}}
        self.pyvista_widget.update_status("multipie", multipie)
        self.misc_button_multipie.pressed.emit()
        self._set_axis_type()

    # ==================================================
    def mp_add_site(self, site, scale=1.0, color=None):
        """
        MultiPie: Add equivalent sites.

        Args:
            site (str): representative site.
            scale (float, optional): size scale.
            color (str, optional): color.
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.object_edit_site.setText(site)
        self.multipie_dialog.dialog.obj_add_site(scale, color)

    # ==================================================
    def mp_add_bond(self, bond, scale=1.0, color=None, color2=None):
        """
        MultiPie: Add equivalent bonds.

        Args:
            bond (str): representative bond.
            scale (float, optional): width scale.
            color (str, optional): color.
            color2 (str, optional): color2.
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.object_edit_bond.setText(bond)
        self.multipie_dialog.dialog.obj_add_bond(scale, color, color2)

    # ==================================================
    def mp_add_vector(self, type, vector, site_bond, scale=1.0):
        """
        MultiPie: Add vectors at equivalent sites or bonds.

        Args:
            type (str): type of vector, Q/G/T/M.
            vector (str): vector (cartesian).
            site_bond (str): representative site or bond.
            scale (float, optional): length scale.
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.object_combo_vector_type.setCurrentText(type)
        self.multipie_dialog.dialog.object_edit_vector.setText(vector + " # " + site_bond)
        self.multipie_dialog.dialog.obj_add_vector(scale)

    # ==================================================
    def mp_add_orbital(self, type, orbital, site_bond, scale=1.0):
        """
        MultiPie: Add orbitals at equivalent sites or bonds.

        Args:
            type (str): type of orbital, Q/G/T/M.
            orbital (str): orbital in terms of x,y,z,r (cartesian).
            site_bond (str): representative site or bond.
            scale (float, optional): size scale.
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.object_combo_orbital_type.setCurrentText(type)
        self.multipie_dialog.dialog.object_edit_orbital.setText(orbital + " # " + site_bond)
        self.multipie_dialog.dialog.obj_add_orbital(scale)

    # ==================================================
    def mp_create_harmonics(self, type, rank):
        """
        MultiPie: Create harmonics list.

        Args:
            type (str): type of harmonics, Q/G/T/M.
            rank (int or str): rank.

        Returns:
            - (list) -- list of harmonics, [str].
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.object_combo_harmonics_type.setCurrentText(type)
        self.multipie_dialog.dialog.object_combo_harmonics_rank.setCurrentText(str(rank))

        lst = self.multipie_dialog.dialog.object_combo_harmonics_irrep.get_item()

        return lst

    # ==================================================
    def mp_add_harmonics(self, tag, site_bond, scale=1.0):
        """
        MultiPie: Add harmonics at equivalent sites or bonds.

        Args:
            tag (str): harmonics tag, obtained by mp_create_harmonics.
            site_bond (str): representative site or bond.
            scale (float, optional): size scale.
        """
        self.multipie_dialog.dialog.object_combo_harmonics_irrep.setCurrentText(tag)
        self.multipie_dialog.dialog.object_edit_harmonics.setText(site_bond)
        self.multipie_dialog.dialog.obj_add_harmonics(scale)

    # ==================================================
    def mp_create_site_samb(self, site):
        """
        MultiPie: Create site SAMB.

        Args:
            site (str): representative site.

        Returns:
            - (list) -- list of site SAMB, [str].
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.basis_edit_site.setText(site)
        d = self.multipie_dialog.dialog.basis_gen_site()

        return d

    # ==================================================
    def mp_add_site_samb(self, tag, scale=1.0):
        """
        MultiPie: Add site SAMB.

        Args:
            tag (str): site SAMB, obtained by mp_create_site_samb.
            scale (float, optional): size scale.
        """
        self.multipie_dialog.dialog.basis_combo_site_samb.setCurrentText(tag)
        self.multipie_dialog.dialog.basis_add_site(scale)

    # ==================================================
    def mp_create_bond_samb(self, bond):
        """
        MultiPie: Create bond SAMB.

        Args:
            bond (str): representative bond.

        Returns:
            - (list) -- list of bond SAMB, [str].
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.basis_edit_bond.setText(bond)
        d = self.multipie_dialog.dialog.basis_gen_bond()

        return d

    # ==================================================
    def mp_add_bond_samb(self, tag, scale=1.0):
        """
        MultiPie: Add bond SAMB.

        Args:
            tag (str): bond SAMB, obtained by mp_create_bond_samb.
            scale (float, optional): width scale.
        """
        self.multipie_dialog.dialog.basis_combo_bond_samb.setCurrentText(tag)
        self.multipie_dialog.dialog.basis_add_bond(scale)

    # ==================================================
    def mp_create_vector_samb(self, type, site_bond):
        """
        MultiPie: Create vector SAMB.

        Args:
            type (str): type of vector, Q/G/T/M.
            site_bond (str): representative site or bond.

        Returns:
            - (list) -- list of vector SAMB, [str].
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.basis_combo_vector_type.setCurrentText(type)
        self.multipie_dialog.dialog.basis_edit_vector.setText(site_bond)
        self.multipie_dialog.dialog.basis_gen_vector()

        z_samb = self.multipie_dialog.dialog.plus["vector_z_samb"]
        d = []
        for select in z_samb.values():
            d += [f"{i[0][0]}{no+1:02d}: {i[0]}" for no, i in enumerate(select)]

        return d

    # ==================================================
    def mp_add_vector_samb(self, lc, scale=1.0):
        """
        MultiPie: Add vector SAMB.

        Args:
            lc (str): linear combination of vector SAMBs, obtained by mp_create_vector_samb.
            scale (float, optional): length scale.
        """
        num = {k: lc.count(k) for k in ["Q", "G", "T", "M"]}
        z_type = "Q" if num["Q"] + num["G"] > 0 else "T"
        self.multipie_dialog.dialog.basis_combo_vector_samb_type.setCurrentText(z_type)
        self.multipie_dialog.dialog.basis_edit_vector_lc.setText(lc)
        self.multipie_dialog.dialog.basis_add_vector_lc(scale)

    # ==================================================
    def mp_add_vector_samb_modulation(self, mod_list):
        """
        MultiPie: Add vector SAMB with modulation.

        Args:
            mod_list (str): modulation list, "[[tag, coeff, k_vector, cos/sin]]".
        """
        num = {k: mod_list.count(k) for k in ["Q", "G", "T", "M"]}
        z_type = "Q,G" if num["Q"] + num["G"] > 0 else "T,M"
        self.multipie_dialog.dialog.basis_combo_vector_modulation_type.setCurrentText(z_type)
        self.multipie_dialog.dialog.basis_edit_vector_modulation.setText(mod_list)
        self.multipie_dialog.dialog.basis_gen_vector_modulation()
        self.multipie_dialog.dialog._vector_modulation_dialog.accept()

    # ==================================================
    def mp_create_orbital_samb(self, type, rank, site_bond):
        """
        MultiPie: Create orbital SAMB.

        Args:
            type (str): type of orbital, Q/G/T/M.
            rank (int or str): rank.
            site_bond (str): representative site or bond.

        Returns:
            - (list) -- list of orbital SAMB, [str].
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.basis_combo_orbital_type.setCurrentText(type)
        self.multipie_dialog.dialog.basis_combo_orbital_rank.setCurrentText(str(rank))
        self.multipie_dialog.dialog.basis_edit_orbital.setText(site_bond)
        self.multipie_dialog.dialog.basis_gen_orbital()

        z_samb = self.multipie_dialog.dialog.plus["orbital_z_samb"]
        d = []
        for select in z_samb.values():
            d += [f"{i[0][0]}{no+1:02d}: {i[0]}" for no, i in enumerate(select)]

        return d

    # ==================================================
    def mp_add_orbital_samb(self, lc, scale=1.0):
        """
        MultiPie: Add orbital SAMB.

        Args:
            lc (str): linear combination of orbital SAMBs, obtained by mp_create_orbital_samb.
            scale (float, optional): size scale.
        """
        num = {k: lc.count(k) for k in ["Q", "G", "T", "M"]}
        z_type = "Q" if num["Q"] + num["G"] > 0 else "T"
        self.multipie_dialog.dialog.basis_combo_orbital_samb_type.setCurrentText(z_type)
        self.multipie_dialog.dialog.basis_edit_orbital_lc.setText(lc)
        self.multipie_dialog.dialog.basis_add_orbital_lc(scale)

    # ==================================================
    def mp_add_orbital_samb_modulation(self, mod_list):
        """
        MultiPie: Add orbital SAMB with modulation.

        Args:
            mod_list (str): modulation list, "[[tag, coeff, k_vector, cos/sin]]".
        """
        num = {k: mod_list.count(k) for k in ["Q", "G", "T", "M"]}
        z_type = "Q,G" if num["Q"] + num["G"] > 0 else "T,M"
        self.multipie_dialog.dialog.basis_combo_orbital_modulation_type.setCurrentText(z_type)
        self.multipie_dialog.dialog.basis_edit_orbital_modulation.setText(mod_list)
        self.multipie_dialog.dialog.basis_gen_orbital_modulation()
        self.multipie_dialog.dialog._orbital_modulation_dialog.accept()

    # ==================================================
    def mp_add_hopping(self, bond, scale=1.0):
        """
        MultiPie: Add hopping bond directions.

        Args:
            bond (str): representative bond.
            scale (float, optional): length scale.
        """
        if self.multipie_dialog is None:
            self.mp_set_group()

        self.multipie_dialog.dialog.basis_edit_hopping.setText(bond)
        self.multipie_dialog.dialog.basis_add_hopping(scale)



"""
Preference dialog.

This module provides preference dialog for PyVistaWidget.
"""

from PySide6.QtWidgets import QDialog, QTabWidget, QWidget, QDialogButtonBox
from PySide6.QtCore import Qt
from qtdraw.widget.custom_widget import Layout, Label, Combo, Spin, DSpin, Check, VSpacer, HSpacer, ColorSelector
from qtdraw.util.util import create_style_sheet


# ==================================================
class PreferenceDialog(QDialog):
    # ==================================================
    def __init__(self, widget, parent=None):
        """
        Preference dialog.

        Args:
            widget (PyVistaWidget): widget.
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)
        self.setWindowTitle("Preferences")
        self.resize(400, 300)

        self.widget = widget
        self.preference = widget._preference

        panel_label = self.create_label_panel(self)
        panel_axis = self.create_axis_panel(self)
        panel_cell = self.create_cell_panel(self)
        panel_light = self.create_light_panel(self)
        panel_latex = self.create_latex_panel(self)
        panel_general = self.create_general_panel(self)

        # tab content.
        tab = QTabWidget(self)
        tab.addTab(panel_label, "Label")
        tab.addTab(panel_axis, "Axis")
        tab.addTab(panel_cell, "Cell")
        tab.addTab(panel_light, "Light")
        tab.addTab(panel_latex, "LaTeX")
        tab.addTab(panel_general, "General")

        # button.
        button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply)
        button.accepted.connect(self.accept)
        button.rejected.connect(self.reject)
        button.button(QDialogButtonBox.Apply).clicked.connect(self.apply)

        # main layout.
        layout = Layout(self)
        layout.setContentsMargins(10, 5, 10, 5)
        layout.addWidget(tab, 0, 0, 1, 2)
        layout.addWidget(button, 1, 0, 2, 1)

    # ==================================================
    def create_label_panel(self, parent):
        """
        Create label panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["label"]

        # widgets.
        label_font = Label(parent, "font")
        combo_font = Combo(parent, ["arial", "courier", "times"])
        label_size = Label(parent, "size")
        spin_size = Spin(parent, 4, 48)
        label_color = Label(parent, "color")
        combo_color = ColorSelector(parent, preference["color"], color_type="color")
        check_bold = Check(parent, "bold")
        check_italic = Check(parent, "italic")
        check_default = Check(parent, "default check")

        # set layout.
        layout.addWidget(label_font, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_font, 0, 1, 1, 1)
        layout.addWidget(label_size, 0, 2, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_size, 0, 3, 1, 1)
        layout.addWidget(check_bold, 1, 1, 1, 1)
        layout.addWidget(check_italic, 1, 3, 1, 1)
        layout.addWidget(label_color, 2, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_color, 2, 1, 1, 1)
        layout.addWidget(check_default, 3, 1, 1, 1)
        layout.addItem(HSpacer(), 0, 4, 1, 1)
        layout.addItem(VSpacer(), 4, 0, 1, 1)

        # initial values.
        combo_font.setCurrentText(preference["font"])
        spin_size.setProperty("value", preference["size"])
        check_bold.setChecked(preference["bold"])
        check_italic.setChecked(preference["italic"])
        check_default.setChecked(preference["default_check"])

        # connections.
        combo_font.currentTextChanged.connect(lambda v: preference.update({"font": v}))
        spin_size.valueChanged.connect(lambda v: preference.update({"size": v}))
        combo_color.currentTextChanged.connect(lambda v: preference.update({"color": v}))
        check_bold.toggled.connect(lambda flag: preference.update({"bold": flag}))
        check_italic.toggled.connect(lambda flag: preference.update({"italic": flag}))
        check_default.toggled.connect(lambda flag: preference.update({"default_check": flag}))

        return panel

    # ==================================================
    def create_axis_panel(self, parent):
        """
        Create axis panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["axis"]
        axis_type = {
            "xyz": "[x,y,z]",
            "abc": "[a,b,c]",
            "abc*": "[a*,b*,c*]",
            "[x,y,z]": "xyz",
            "[a,b,c]": "abc",
            "[a*,b*,c*]": "abc*",
        }

        # widgets.
        label_type = Label(parent, "type")
        combo_type = Combo(parent, ["xyz", "abc", "abc*"])
        label_size = Label(parent, "size")
        spin_size = Spin(parent, 12, 36)
        check_bold = Check(parent, "bold")
        check_italic = Check(parent, "italic")

        # set layout.
        layout.addWidget(label_type, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_type, 0, 1, 1, 1)
        layout.addWidget(label_size, 0, 2, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_size, 0, 3, 1, 1)
        layout.addWidget(check_bold, 1, 1, 1, 1)
        layout.addWidget(check_italic, 1, 3, 1, 1)
        layout.addItem(HSpacer(), 0, 4, 1, 1)
        layout.addItem(VSpacer(), 2, 0, 1, 1)

        # initial values.
        combo_type.setCurrentText(axis_type[preference["label"]])
        spin_size.setProperty("value", preference["size"])
        check_bold.setChecked(preference["bold"])
        check_italic.setChecked(preference["italic"])

        # connections.
        combo_type.currentTextChanged.connect(lambda v: preference.update({"label": axis_type[v]}))
        spin_size.valueChanged.connect(lambda v: preference.update({"size": v}))
        check_bold.toggled.connect(lambda flag: preference.update({"bold": flag}))
        check_italic.toggled.connect(lambda flag: preference.update({"italic": flag}))

        return panel

    # ==================================================
    def create_cell_panel(self, parent):
        """
        Create cell panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["cell"]

        # widgets.
        label_width = Label(parent, "width")
        spin_width = DSpin(parent, 0.0, 5.0, 0.1)
        label_color = Label(parent, "color")
        combo_color = ColorSelector(parent, preference["color"], color_type="color")
        label_opacity = Label(parent, "opacity")
        spin_opacity = DSpin(parent, 0.0, 1.0, 0.1)

        # set layout.
        layout.addWidget(label_width, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_width, 0, 1, 1, 1)
        layout.addWidget(label_opacity, 1, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_opacity, 1, 1, 1, 1)
        layout.addWidget(label_color, 2, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_color, 2, 1, 1, 2)
        layout.addItem(HSpacer(), 2, 3, 1, 1)
        layout.addItem(VSpacer(), 3, 0, 1, 1)

        # initial values.
        spin_width.setProperty("value", preference["line_width"])
        spin_opacity.setProperty("value", preference["opacity"])

        # connections.
        spin_width.valueChanged.connect(lambda v: preference.update({"line_width": round(v, 4)}))
        combo_color.currentTextChanged.connect(lambda v: preference.update({"color": v}))
        spin_opacity.valueChanged.connect(lambda v: preference.update({"opacity": round(v, 4)}))

        return panel

    # ==================================================
    def create_light_panel(self, parent):
        """
        Create light panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["light"]

        # widgets.
        label_intensity = Label(parent, "intensity")
        spin_intensity = DSpin(parent, 0.0, 1.0, 0.05)
        check_pbr = Check(parent, "physics based rendering")
        label_metallic = Label(parent, "metallic")
        spin_metallic = DSpin(parent, 0.0, 1.0, 0.1)
        label_roughness = Label(parent, "roughness")
        spin_roughness = DSpin(parent, 0.0, 1.0, 0.1)
        label_color = Label(parent, "color")
        combo_color = ColorSelector(parent, preference["color"], color_type="color")
        label_type = Label(parent, "type")
        combo_type = Combo(parent, ["lightkit", "3 lights", "ver1"])

        # set layout.
        layout.addWidget(label_type, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_type, 0, 1, 1, 1)
        layout.addWidget(label_intensity, 1, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_intensity, 1, 1, 1, 1)
        layout.addWidget(label_color, 1, 2, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_color, 1, 3, 1, 1)
        layout.addWidget(check_pbr, 2, 0, 1, 3)
        layout.addWidget(label_metallic, 3, 1, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_metallic, 3, 2, 1, 1)
        layout.addWidget(label_roughness, 4, 1, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_roughness, 4, 2, 1, 1)
        layout.addItem(HSpacer(), 0, 4, 1, 1)
        layout.addItem(VSpacer(), 5, 0, 1, 1)

        # initial values.
        spin_intensity.setProperty("value", preference["intensity"])
        check_pbr.setChecked(preference["pbr"])
        spin_metallic.setProperty("value", preference["metallic"])
        spin_roughness.setProperty("value", preference["roughness"])
        combo_type.setCurrentText(preference["type"])

        # connections.
        spin_intensity.valueChanged.connect(lambda v: preference.update({"intensity": round(v, 4)}))
        check_pbr.toggled.connect(lambda flag: preference.update({"pbr": flag}))
        spin_metallic.valueChanged.connect(lambda v: preference.update({"metallic": round(v, 4)}))
        spin_roughness.valueChanged.connect(lambda v: preference.update({"roughness": round(v, 4)}))
        combo_color.currentTextChanged.connect(lambda v: preference.update({"color": v}))
        combo_type.currentTextChanged.connect(lambda v: preference.update({"type": v}))

        return panel

    # ==================================================
    def create_latex_panel(self, parent):
        """
        Create LaTeX panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["latex"]

        # widgets.
        label_color = Label(parent, "color")
        combo_color = ColorSelector(parent, preference["color"], color_type="color")
        label_size = Label(parent, "size")
        spin_size = Spin(parent, 8, 24)
        label_dpi = Label(parent, "DPI")
        combo_dpi = Combo(parent, ["120", "240", "300"])

        # set layout.
        layout.addWidget(label_color, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_color, 0, 1, 1, 2)
        layout.addWidget(label_size, 1, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_size, 1, 1, 1, 2)
        layout.addWidget(label_dpi, 2, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_dpi, 2, 1, 1, 2)
        layout.addItem(HSpacer(), 0, 3, 1, 1)
        layout.addItem(VSpacer(), 3, 0, 1, 1)

        # initial values.
        spin_size.setProperty("value", preference["size"])
        combo_dpi.setCurrentText(str(preference["dpi"]))

        # connections.
        combo_color.currentTextChanged.connect(lambda v: preference.update({"color": v}))
        spin_size.valueChanged.connect(lambda v: preference.update({"size": v}))
        combo_dpi.currentTextChanged.connect(lambda v: preference.update({"dpi": int(v)}))

        return panel

    # ==================================================
    def create_general_panel(self, parent):
        """
        Create general panel.
        """
        panel = QWidget(parent)
        layout = Layout(panel)
        layout.setContentsMargins(10, 5, 10, 5)

        preference = self.preference["general"]

        # widgets.
        label_style = Label(parent, "style")
        combo_style = Combo(parent, ["fusion", "macos", "windows"])
        label_font = Label(parent, "font")
        combo_font = Combo(parent, ["Osaka", "Monaco", "Arial", "Times New Roman", "Helvetica Neue"])
        label_color = Label(parent, "scheme")
        combo_color = Combo(parent, ["Jmol", "VESTA"])
        label_size = Label(parent, "size")
        spin_size = Spin(parent, 9, 14)

        # set layout.
        layout.addWidget(label_style, 0, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_style, 0, 1, 1, 1)
        layout.addWidget(label_font, 1, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_font, 1, 1, 1, 1)
        layout.addWidget(label_size, 1, 2, 1, 1, Qt.AlignRight)
        layout.addWidget(spin_size, 1, 3, 1, 1)
        layout.addWidget(label_color, 2, 0, 1, 1, Qt.AlignRight)
        layout.addWidget(combo_color, 2, 1, 1, 1)
        layout.addItem(HSpacer(), 1, 4, 1, 1)
        layout.addItem(VSpacer(), 3, 0, 1, 1)

        # initial values.
        combo_style.setCurrentText(preference["style"])
        combo_font.setCurrentText(preference["font"])
        combo_color.setCurrentText(preference["color_scheme"])
        spin_size.setProperty("value", preference["size"])

        # connections.
        combo_style.currentTextChanged.connect(lambda v: preference.update({"style": v}))
        combo_font.currentTextChanged.connect(lambda v: preference.update({"font": v}))
        combo_color.currentTextChanged.connect(lambda v: preference.update({"color_scheme": v}))
        spin_size.valueChanged.connect(lambda v: preference.update({"size": v}))

        return panel

    # ==================================================
    def apply(self, button):
        """
        Apply change.

        Args:
            button (QAbstractButton): button status.

        :meta private:
        """
        self.parent().app.setStyleSheet(create_style_sheet(self.preference["general"]["size"]))
        self.widget.refresh()
        self.widget.redraw()
        self.parent()._update_panel()



"""
Delegate for GroupModel and GroupView.

This module provides delegate for color selector, combo, and editor.
"""

from PySide6.QtWidgets import QStyle, QStyledItemDelegate
from PySide6.QtCore import Qt
from qtdraw.widget.custom_widget import Combo, Editor, ColorSelector


# ==================================================
class Delegate(QStyledItemDelegate):
    # ==================================================
    def __init__(self, parent):
        """
        Base delegate.

        Args:
            parent (QWidget): parent.
        """
        super().__init__(parent)

    # ==================================================
    def setEditorData(self, editor, index):
        """
        Set editor data.

        Args:
            editor (QWidget): editor.
            index (QModelIndex): index.
        """
        value = index.model().data(index, Qt.EditRole)
        editor.setCurrentText(value)

    # ==================================================
    def setModelData(self, editor, model, index):
        return  # set model data by using signal.

    # ==================================================
    def updateEditorGeometry(self, editor, option, index):
        """
        Update editor geometry.

        Args:
            editor (QWidget): editor.
            option (QStyleOptionViewItem) option.
            index (QModelIndex): index.
        """
        editor.setGeometry(option.rect.adjusted(5, 7, -5, 0))

    # ==================================================
    def paint(self, painter, option, index):
        """
        Paint cell (need to highlight selection).

        Args:
            painter (QPainter): painter.
            option (QStyleOptionViewItem) option.
            index (QModelIndex): index.
        """
        if index.isValid():
            value = index.model().data(index, Qt.DisplayRole)
            if value:
                option.displayAlignment = Qt.AlignHCenter | Qt.AlignVCenter
                self.parent().style().drawControl(QStyle.CE_ItemViewItem, option, painter, self.parent())


# ==================================================
class ComboDelegate(Delegate):
    # ==================================================
    def __init__(self, parent, default, option):
        """
        Create delegate for Combo.

        Args:
            parent (QWidget): parent.
            default (str): Combo default.
            option (tuple): Cobo option.
        """
        super().__init__(parent)
        self.default = default
        self.option = option

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create combobox.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (Combo) -- combo widget.
        """
        combo = Combo(parent, self.option, self.default)
        combo.currentTextChanged.connect(lambda data: index.model().setData(index, data))
        return combo


# ==================================================
class ColorDelegate(Delegate):
    # ==================================================
    def __init__(self, parent, default, option):
        """
        Create delegate for ColorSelector.

        Args:
            parent (QWidget): parent.
            default (str): ColorSelector default.
            option (tuple): ColorSelector option.
        """
        super().__init__(parent)
        self.default = default
        self.option = option

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create color selector.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (ColorSelector) -- color selector widget.
        """
        color_selector = ColorSelector(parent, self.default, self.option)
        color_selector.currentTextChanged.connect(lambda data: index.model().setData(index, data))

        return color_selector


# ==================================================
class EditorDelegate(Delegate):
    # ==================================================
    def __init__(self, parent, default, option, t, color, size, dpi):
        """
        Create delegate for Editor.

        Args:
            parent (QWidget): parent.
            default (str): Editor default.
            option (tuple): Editor option.
            t (str): Editor type.
            color (str): Editor color.
            size (int): Editor size.
            dpi (int): Editor dpi.
        """
        super().__init__(parent)
        self.default = default
        self.option = option
        self.type = t
        self.color = color
        self.size = size
        self.dpi = dpi

    # ==================================================
    def createEditor(self, parent, option, index):
        """
        Create editor.

        Args:
            parent (QWidget): parent.
            option (QStyleOptionViewItem): option.
            index (QModelIndex): index.

        Returns:
            - (Editor) -- editor widget.
        """
        editor = Editor(parent, self.default, (self.type, self.option), self.color, self.size, self.dpi)

        editor.returnPressed.connect(lambda data: index.model().setData(index, data))
        return editor

    # ==================================================
    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect.adjusted(5, 0, -5, 0))



"""
Utility for parsing material files.

This module contains utilities for parsing.
"""

import numpy as np
import copy
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import MinimumDistanceNN
from pymatgen.io.cif import CifParser
from qtdraw.core.pyvista_widget_setting import DIGIT, default_preference, default_status
from qtdraw.parser.element import element_color
from qtdraw.core.pyvista_widget_setting import widget_detail as detail
from qtdraw.parser.vesta import parse_vesta, create_structure_vesta
from qtdraw.parser.data_group import _data_no_space_group
from qtdraw import __version__


# ==================================================
def get_model_cell(graph):
    """
    Get model and cell.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns:
        - (str) -- name of model.
        - (dict) -- unit-cell info.
    """
    name = graph.structure.composition.reduced_formula

    lat = graph.structure.lattice
    cell = {
        "a": np.round(lat.a, DIGIT),
        "b": np.round(lat.b, DIGIT),
        "c": np.round(lat.c, DIGIT),
        "alpha": np.round(lat.alpha, DIGIT),
        "beta": np.round(lat.beta, DIGIT),
        "gamma": np.round(lat.gamma, DIGIT),
    }

    return name, cell


# ==================================================
def get_site_info(graph):
    """
    Get site information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns:
        - (list) -- site_info. (name, label, element, frac_coords, radius).
    """
    structure = graph.structure
    eq_sites = structure.equivalent_sites

    # grouping equivalent sites.
    MIN_RADIUS = 0.25

    site_info = []

    element0 = ""
    group_index = 0
    for group in eq_sites:
        group_index += 1
        for site_index, site in enumerate(group):
            element = site.species.elements[0].symbol
            if element != element0:
                element0 = element
                group_index = 0
            symbols = ''.join([el.symbol for el in site.species.elements])
            name = f"{symbols}{group_index+1}"
            label = f"{name}_{site_index+1}"
            frac_coords = site.frac_coords

            radius = 0.0
            for el, occu in site.species.items():
                r = el.atomic_radius or MIN_RADIUS
                radius += occu * r

            site_info.append((
                name,
                label,
                element,
                frac_coords.round(DIGIT),
                round(radius,DIGIT-2)
            ))

    return site_info


# ==================================================
def get_bond_info(graph, site_info):
    """
    Get bond information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.
        site_info (list): site info.

    Returns:
        - (list) -- bond_info. (name, label, center, vector, tail_element, head_element).
    """
    adjacency = graph.as_dict()["graphs"]["adjacency"]

    dbonds = {}
    for tail, tail_adjacency in zip(site_info, adjacency):
        t_name, _, t_element, t_pos, _ = tail
        for head_info in tail_adjacency:
            h_name, _, h_element, h_pos, _ = site_info[head_info["id"]]
            h_pos = np.array(head_info["to_jimage"]) + h_pos
            center = ((t_pos + h_pos) / 2).round(DIGIT)
            vector = (h_pos - t_pos).round(DIGIT)
            key = t_name + "-" + h_name
            dbonds[key] = dbonds.get(key, []) + [(center, vector, t_element, h_element)]

    bond_info = []
    for name, v in dbonds.items():
        for no, (center, vector, t_element, h_element) in enumerate(v):
            label = name + "_" + str(no + 1)
            bond_info.append((name, label, center, vector, t_element, h_element))

    return bond_info


# ==================================================
def draw_site_bond(widget, name, site_info, bond_info):
    """
    Draw site and bond.

    Args:
        widget (PyVistaWidget): widget.
        name (str): name of model.
        site_info (list): site info.
        bond_info (list): bond info.
    """
    site_scale = detail["site_scale"]
    bond_scale = detail["bond_scale"]
    color_scheme = widget._preference["general"]["color_scheme"]
    default_color = "silver"

    # draw sites.
    widget._data["site"].block_update_widget(True)
    for name, label, element, frac_coords, radius in site_info:
        color = element_color[color_scheme].get(element, default_color)
        widget.add_site(position=str(frac_coords.tolist()), name=name, label=label, color=color, size=site_scale * radius)
    widget._data["site"].block_update_widget(False)

    # draw bonds.
    widget._data["bond"].block_update_widget(True)
    for name, label, center, vector, tail_element, head_element in bond_info:
        tail_color = element_color[color_scheme].get(tail_element, default_color)
        head_color = element_color[color_scheme].get(head_element, default_color)
        widget.add_bond(
            position=str(center.tolist()),
            direction=str(vector.tolist()),
            color=tail_color,
            color2=head_color,
            name=name,
            label=label,
            width=bond_scale,
        )
    widget._data["bond"].block_update_widget(False)


# ==================================================
def parse_material(filename):
    """
    Parse material file.

    Args:
        filename (str): filename.

    Returns:
        - (dict) -- data for PyVistaWidget.
        - (list) -- site info. to draw.
        - (list) -- bond info. to draw.
        - (Structure) -- symmetrized structure.
    """
    if filename.endswith(".vesta"):
        vesta_dict = parse_vesta(filename)
        structure = create_structure_vesta(vesta_dict)
    elif filename.endswith(".cif"):
        parser = CifParser(filename)
        structure = parser.parse_structures(primitive=False, symmetrized=False)[0]
    elif filename.endswith(".xsf"):
        structure = Structure.from_file(filename)
        sga = SpacegroupAnalyzer(structure)
        structure = sga.get_refined_structure()
    else:
        structure = Structure.from_file(filename)

    sga = SpacegroupAnalyzer(structure)
    symmetrized = sga.get_symmetrized_structure()
    sg_no = sga.get_space_group_number()
    env = MinimumDistanceNN()
    graph = StructureGraph.from_local_env_strategy(symmetrized, env)

    name, cell = get_model_cell(graph)
    crystal = sga.get_crystal_system()

    group_name = _data_no_space_group[sg_no]
    multipie = {"group": {"group": group_name}}

    status = copy.deepcopy(default_status)
    status.update({"model": name, "crystal": crystal, "cell": cell, "clip": False, "multipie": multipie})
    preference = copy.deepcopy(default_preference)
    preference["axis"]["label"] = "[a,b,c]"
    preference["label"]["default_check"] = False
    all_data = {"version": __version__, "status": status, "preference": preference}

    site_info = get_site_info(graph)
    bond_info = get_bond_info(graph, site_info)

    return all_data, site_info, bond_info, symmetrized



"""
Control Qt event loop.

This module provides functions to control Qt event loop.

See also,

- [use Qt from Jupyter NB](https://qiita.com/Hanjin_Liu/items/9df684920727f8a784c4)
- [exception hook](https://timlehr.com/python-exception-hooks-with-qt-message-box/)
"""

import sys
import os
import logging
import asyncio
from IPython.core import ultratb
from PySide6.QtCore import QObject, Signal, Qt, QLoggingCategory
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QApplication
from qtdraw.widget.message_box import MessageBox
from qtdraw.util.util import set_latex_setting, create_style_sheet
from qtdraw.core.pyvista_widget_setting import default_preference
from qtdraw import __top_dir__


# ==================================================
def gui_qt():
    """
    Execute Qt GUI mode in IPython (if available).
    """
    try:
        from IPython import get_ipython
    except ImportError:
        get_ipython = lambda: False

    shell = get_ipython()

    if shell and getattr(shell, "enable_gui", None):
        if shell.active_eventloop != "qt":
            shell.enable_gui("qt")


# ==================================================
def get_qt_application(latex=True):
    """
    Get Qt application.

    Args:
        latex (bool, optional): use latex setting ?

    Returns:
        - (QApplication) -- Qt application.

    Note:
        - call this before super().__init__() in __init__() of QMainWindow class as follows.
            - self.app = get_qt_application()
            - ExceptionHook()
    """
    # suppress logging message concerning japanese IM.
    QLoggingCategory.setFilterRules("qt.qpa.keymapper=false")

    # initial LaTeX setting.
    if latex:
        set_latex_setting()

    gui_qt()  # does nothing in Jupyter7+.
    app = QApplication.instance()

    if app is None:
        app = QApplication([])

    # for high-resolution setting.
    app.setAttribute(Qt.AA_EnableHighDpiScaling)
    app.setAttribute(Qt.AA_UseHighDpiPixmaps)

    # set general appearance.
    style = default_preference["general"]["style"]
    font = default_preference["general"]["font"]
    size = default_preference["general"]["size"]
    app.setStyle(style)
    app.setFont(QFont(font))
    app.setStyleSheet(create_style_sheet(size))

    # integrate qt event loop into asyncio for Jupyter environment.
    in_jupyter = "JPY_PARENT_PID" in os.environ or "ipykernel" in sys.modules
    if in_jupyter:

        async def _keep_qt_alive():
            while True:
                app.processEvents()
                await asyncio.sleep(0.01)

        # avoid double-scheduling.
        loop = asyncio.get_event_loop()
        existing = [t for t in asyncio.all_tasks(loop) if t._coro.__name__ == "_keep_qt_alive"]
        if not existing:
            asyncio.create_task(_keep_qt_alive())

    return app


# ==================================================
class ExceptionHook(QObject):
    msg_signal = Signal(str)

    # ==================================================
    def __init__(self, parent=None):
        """
        Exception hook.

        Args:
            parent (QObject, optional): parent.

        Note:
            - call this before super().__init__() in __init__() of QMainWindow class as follows.
                - self.app = get_qt_application()
                - ExceptionHook()
        """
        super().__init__(parent)

        # hook the exception handler of the Python interpreter.
        sys.excepthook = self.hook

        # connection.
        self.msg_signal.connect(lambda x: MessageBox(x, "Exception Message"))

    # ==================================================
    def hook(self, type, value, traceback):
        """
        Callback for uncaught exceptions.

        Args:
            type (Any): type of exception.
            value (Any): arguments.
            traceback (TraceBack): traceback.
        """
        if issubclass(type, KeyboardInterrupt):
            sys.__excepthook__(type, value, traceback)  # ignore keyboard interrupt for console applications.
        else:
            bar = "---------------------------------------------------------------------------"
            handler = ultratb.VerboseTB(color_scheme="NoColor", long_header=False)
            log_msg = handler.text(type, value, traceback)
            log_msg += "\n" + bar
            simple = "\n" + bar + "\n" + ultratb.SyntaxTB(theme_name="NoColor").text(type, value, traceback) + bar
            self.msg_signal.emit(log_msg)
            logging.critical(simple)

    # ==================================================
    def reset(self):
        """
        Reset exception hook.
        """
        sys.excepthook = sys.__excepthook__



"""
View panel for GroupModel.

This module provides a view class to show object data.
By clicking right button of mouse, the context menu appears.
"""

from PySide6.QtWidgets import QMenu, QTreeView, QHeaderView
from PySide6.QtCore import Qt, Signal, QPoint, QModelIndex, QItemSelection, QItemSelectionModel, QObject, QEvent
from qtdraw.core.pyvista_widget_setting import COLOR_WIDGET, COMBO_WIDGET, EDITOR_WIDGET, HIDE_TYPE
from qtdraw.widget.delegate import ColorDelegate, ComboDelegate, EditorDelegate


# ==================================================
class GroupView(QTreeView):
    selectionChanged = Signal(str, list, list)  # name, deselect, select.

    # ==================================================
    def __init__(self, model, parent=None, use_delegate=True):
        """
        Group view.

        Args:
            model (GroupModel): group model.
            parent (QWidget, optional): parent.
        """
        super().__init__(parent)

        # for debug.
        self._debug = {"delegate": use_delegate, "hide": True, "raw_data": False}

        # set model.
        self.setModel(model)

        # set delegate.
        if self._debug["delegate"]:
            for c, t in enumerate(self.model().column_type):
                if t in COLOR_WIDGET:
                    default = self.model().column_default[c]
                    option = t
                    self.setItemDelegateForColumn(c, ColorDelegate(self, default, option))
                elif t in COMBO_WIDGET:
                    default = self.model().column_default[c]
                    option = self.model().column_option[c]
                    self.setItemDelegateForColumn(c, ComboDelegate(self, default, option))
                elif t in EDITOR_WIDGET:
                    default = self.model().column_default[c]
                    option = self.model().column_option[c]
                    if hasattr(model.parent(), "_preference"):
                        color = model.parent()._preference["latex"]["color"]
                        size = model.parent()._preference["latex"]["size"]
                        dpi = model.parent()._preference["latex"]["dpi"]
                    else:
                        color, size, dpi = "black", 11, 120
                    self.setItemDelegateForColumn(c, EditorDelegate(self, default, option, t, color, size, dpi))

        # set properties.
        self.setAlternatingRowColors(True)
        self.header().setSectionsMovable(False)
        for column in range(self.model().columnCount()):
            self.header().setSectionResizeMode(column, QHeaderView.ResizeToContents)
        self.header().setDefaultAlignment(Qt.AlignHCenter | Qt.AlignVCenter)

        # hide columns.
        if self._debug["hide"]:
            for column, role in enumerate(self.model().column_type):
                if role in HIDE_TYPE:
                    self.header().setSectionHidden(column, True)

        # context menu.
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)

        # connection to update widget.
        if self._debug["delegate"]:
            self.model().updateWidget.connect(self.update_widget)
            self.update_widget(force=True)

        self.selectionModel().selectionChanged.connect(self.selection_changed)

        self.clear_selection()

    # ==================================================
    def clear_selection(self):
        """
        Clear selection.
        """
        self.clearFocus()
        self.clearSelection()
        self.setCurrentIndex(QModelIndex())
        self.setFocus()

    # ==================================================
    def mousePressEvent(self, event):
        """
        Mouse press event for focus or clear selection.
        """
        self.setFocus()
        # in order to deselect when click point has no item.
        position = event.position()
        point = QPoint(int(position.x()), int(position.y()))
        index = self.indexAt(point)
        if not index.isValid():
            self.clear_selection()
        super().mousePressEvent(event)

    # ==================================================
    def keyPressEvent(self, event):
        """
        Key press event for ESC and up and down keys.
        """
        # in order to deselect by ESC key.
        if event.key() == Qt.Key_Escape:
            self.clear_selection()
        # in order to deselect when arrow key selection goes outside.
        elif event.key() in [Qt.Key_Up, Qt.Key_Down]:
            current_indexes = self.selectionModel().selectedIndexes()

            if len(current_indexes) == 0:
                super().keyPressEvent(event)
                return

            first_index = current_indexes[0]
            last_index = current_indexes[-1]

            if event.key() == Qt.Key_Up:
                next_index = self.indexAbove(first_index)
            elif event.key() == Qt.Key_Down:
                next_index = self.indexBelow(last_index)
            else:
                next_index = QModelIndex()

            if not next_index.isValid():
                self.clear_selection()
            else:  # for within range.
                super().keyPressEvent(event)
        else:
            super().keyPressEvent(event)

    # ==================================================
    def context_menu(self, position):
        """
        Context menu.
        """
        index = self.indexAt(position)

        self.menu = QMenu(self)
        self.menu.addAction("insert", self.insert_row)
        if index.isValid():
            self.menu.addAction("copy", self.copy_row)
            self.menu.addAction("remove", self.remove_row)
        if self._debug["raw_data"]:
            self.menu.addAction("raw_data", lambda: print(self.model().tolist()))
        self.menu.exec(self.mapToGlobal(position))

    # ==================================================
    def insert_row(self):
        """
        Insert row.
        """
        indexes = self.selectedIndexes()
        idx = self.model().action_insert_row(indexes)
        if idx.isValid():
            self.expand(idx)

    # ==================================================
    def copy_row(self):
        """
        Copy row.
        """
        indexes = self.selectedIndexes()
        self.model().action_copy_row(indexes)

    # ==================================================
    def remove_row(self):
        """
        Remove row.
        """
        indexes = self.selectedIndexes()
        self.model().action_remove_row(indexes)

    # ==================================================
    def update_widget(self, force=False):
        """
        Update widget.
        """
        if self.isVisible() or force:
            self.blockSignals(True)
            self.setUpdatesEnabled(False)
            self.set_widget(self.model().invisibleRootItem())
            self.setUpdatesEnabled(True)
            self.blockSignals(False)
            self.hide()  # in order to refresh widget.
            self.show()

    # ==================================================
    def set_widget(self, item):
        """
        Set widget.

        Args:
            item (QStandardItem): item.
        """
        for row in range(item.rowCount()):
            child = item.child(row)
            if child:
                for column in self.model().column_widget:
                    index = self.model().index(child.row(), column, self.model().parent(self.model().indexFromItem(child)))
                    self.closePersistentEditor(index)
                    self.openPersistentEditor(index)
                self.set_widget(child)

    # ==================================================
    def selection_changed(self, selected, deselected):
        """
        For Selection changed.

        Args:
            selected (list): selected indexes.
            deselected (list): deselected indexes.
        """
        model = self.model()

        deselected = [i for i in deselected.indexes() if i.column() == 0]
        if deselected:
            index0 = deselected[0]
            row_data0 = []
            if model.rowCount(index0) == 0:
                row_data0.append(model.get_row_data(index0))
            else:
                for row in range(model.rowCount(index0)):
                    cindex0 = model.index(row, 0, index0)
                    row_data0.append(model.get_row_data(cindex0))
        else:
            row_data0 = None

        selected = [i for i in selected.indexes() if i.column() == 0]
        if selected:
            index1 = selected[0]
            row_data1 = []
            if model.rowCount(index1) == 0:
                row_data1.append(model.get_row_data(index1))
            else:
                for row in range(model.rowCount(index1)):
                    cindex1 = model.index(row, 0, index1)
                    row_data1.append(model.get_row_data(cindex1))
        else:
            row_data1 = None

        self.selectionChanged.emit(model.group_name, row_data0, row_data1)

    # ==================================================
    def select_row(self, index):
        """
        Select row.

        Args:
            index (QModelIndex): index.
        """
        index = index.siblingAtColumn(0)
        if index.parent().isValid():  # in case of child.
            self.expand(index.parent())
        index1 = index.siblingAtColumn(self.model().columnCount() - 1)
        selection = QItemSelection(index, index1)  # all columns.
        self.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)



"""
Converter for QtDraw ver. 1.

This module contains the converter for version 1 format.
"""

import copy
from qtdraw.core.pyvista_widget_setting import default_status, default_preference
from qtdraw.multipie.plugin_multipie_setting import space_group_list, point_group_list, crystal_list
from qtdraw.multipie.plugin_multipie_setting import default_status as multipie_default
from qtdraw import __version__


# ==================================================
def to_bool(b):
    """
    Convert from string to bool.

    Args:
        b (str): string of bool.

    Returns:
        - (bool) -- converted value.
    """
    if type(b) == str:
        return b == "True"
    else:
        return b


# old setting keyword (text -> text2d).
# ==================================================
old_setting = {
    "site": ["name", "ON", "label", "cell", "position", "size", "color", "opacity", "space", "D", "AD", "AL"],
    "bond": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "vector",
        "width",
        "color",
        "color2",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "vector": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "vector",
        "length",
        "width",
        "offset",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "orbital": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "shape",
        "surface",
        "size",
        "scale",
        "theta0",
        "theta1",
        "phi0",
        "phi1",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "stream": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "shape",
        "vector",
        "size",
        "v_size",
        "width",
        "scale",
        "theta",
        "phi",
        "theta0",
        "theta1",
        "phi0",
        "phi1",
        "color",
        "component",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "plane": ["name", "ON", "label", "cell", "position", "normal", "x", "y", "color", "opacity", "space", "D", "AD", "AL"],
    "box": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "a1",
        "a2",
        "a3",
        "edge",
        "wireframe",
        "width",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "polygon": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "point",
        "connection",
        "edge",
        "wireframe",
        "width",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "text3d": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "text",
        "size",
        "depth",
        "normal",
        "offset",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "spline": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "point",
        "width",
        "n_interp",
        "closed",
        "natural",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "spline_t": [
        "name",
        "ON",
        "label",
        "cell",
        "position",
        "expression",
        "t_range",
        "width",
        "n_interp",
        "closed",
        "natural",
        "color",
        "opacity",
        "space",
        "D",
        "AD",
        "AL",
    ],
    "caption": ["name", "cell", "position", "caption", "space", "size", "bold", "color", "D", "AD"],
    "text": ["name", "position", "relative", "caption", "size", "color", "font", "D", "AD"],
}


# ==================================================
def get_status(dic):
    """
    Get update status.

    Args:
        dic (dict): read status.

    Returns:
        - (dict) -- updated status.
    """
    axis_type = {0: "on", 1: "axis", 2: "off"}
    cell_mode = {0: "single", 1: "all", 2: "off"}
    setting = dic["setting"]

    status = copy.deepcopy(default_status)
    update_status = {
        "model": setting["model"],
        "origin": setting["origin"],
        "cell": setting["cell"],
        "crystal": setting["crystal"],
        "clip": setting["clip"],
        "repeat": setting["repeat"],
        "lower": setting["view_range"][0],
        "upper": setting["view_range"][1],
        "view": setting["view"],
        "axis_type": axis_type[setting["axis_mode"]],
        "cell_mode": cell_mode[setting["cell_mode"]],
    }
    status.update(update_status)

    return status


# ==================================================
def get_camera(dic):
    """
    Get camera info.

    Args:
        dic (dict): read camera info.

    Returns:
        - (dict) -- updated camera info.
    """
    setting = dic["setting"]
    camera = {
        "position": [2.495208, 2.99401, 1.497604],
        "viewup": [-0.186248, -0.23281, 0.954521],
        "focal_point": [0.5, 0.5, 0.5],
        "angle": 30.0,
        "clipping_range": [1.68102, 5.450085],
        "distance": 3.346065,
        "scale": 0.866025,
    }
    update_camera = {
        "position": list(setting["camera.position"]),
        "viewup": list(setting["camera.up"]),
        "focal_point": list(setting["camera.focus"]),
    }
    camera.update(update_camera)

    return camera


# ==================================================
def get_preference(dic):
    """
    Get update preference.

    Args:
        dic (dict): read preference.

    Returns:
        - (dict) -- updated preference.
    """
    pref = dic["preference"]
    preference = copy.deepcopy(default_preference)
    amode = {"xyz": "[x,y,z]", "abc": "[a,b,c]", "abc*": "[a*,b*,c*]"}
    update_preference = {
        "light": {
            "type": "ver1",
            "intensity": pref["light.intensity"],
            "pbr": pref["light.pbr"],
            "metallic": pref["light.metallic"],
            "roughness": pref["light.roughness"],
        },
        "label": {
            "font": pref["label.font"],
            "size": pref["label.size"],
            "bold": pref["label.bold"],
            "italic": pref["label.italic"],
            "color": pref["label.color"],
        },
        "cell": {"line_width": pref["cell.width"], "color": pref["cell.color"], "opacity": pref["cell.opacity"]},
        "axis": {
            "size": pref["axis.size"],
            "bold": pref["axis.bold"],
            "italic": pref["axis.italic"],
            "label": amode[pref["axis.type"]],
        },
    }
    for key, value in update_preference.items():
        preference[key].update(value)

    return preference


# ==================================================
def get_multipie(dic):
    """
    Get multipie info.

    Args:
        dic (dict): read multipie info.

    Returns:
        - (dict) -- updated multipie info.
    """
    type_str = {0: "Q", 1: "G", 2: "T", 3: "M"}
    type_str1 = {0: "Q,G", 1: "T,M"}

    multipie = copy.deepcopy(multipie_default)
    multipie["plus"] = {}
    if "multipie" in dic.keys():
        old = dic["multipie"]
        multipie["version"] = old["version"]

        # main panel.
        g_type, crystal, group = old["main"]["group"]
        crystal = crystal_list[crystal]
        if g_type == 0:
            tag = space_group_list[crystal][group]
        else:
            tag = point_group_list[crystal][group]
        if tag.count(" ") > 0:
            tag = tag.split(" ")[1]
        multipie["group"]["group"] = tag

        multipie["group"]["irrep1"] = 0
        multipie["group"]["irrep2"] = 0
        multipie["group"]["irrep"] = 0

        multipie["group"]["vc_wyckoff"] = 0

        # tab1.
        multipie["object"]["site"] = old["tab1"]["site"][0]

        multipie["object"]["bond"] = old["tab1"]["bond"][0]

        vector, v_type = old["tab1"]["vector"]
        multipie["object"]["vector"] = vector
        multipie["object"]["vector_type"] = type_str[v_type]

        orbital, o_type = old["tab1"]["orbital"]
        multipie["object"]["orbital"] = orbital
        multipie["object"]["orbital_type"] = type_str[o_type]

        harmonics, h_type, h_rank, _ = old["tab1"]["harmonics"]
        multipie["object"]["harmonics_type"] = type_str[h_type]
        multipie["object"]["harmonics_rank"] = h_rank
        multipie["object"]["harmonics"] = harmonics

        multipie["object"]["wyckoff"] = old["tab1"]["wyckoff"][0]

        # tab2.
        site = old["tab2"]["site"][0]
        multipie["basis"]["site"] = site

        bond = old["tab2"]["bond"][0]
        multipie["basis"]["bond"] = bond

        vector, v_type, v_samb_type, _, v_lc, v_mod_type, v_mod = old["tab2"]["vector"]
        multipie["basis"]["vector_type"] = type_str[v_type]
        multipie["basis"]["vector"] = vector
        multipie["basis"]["vector_samb_type"] = type_str[v_samb_type]
        multipie["basis"]["vector_lc"] = v_lc
        multipie["basis"]["vector_modulation_type"] = type_str1[v_mod_type]
        multipie["basis"]["vector_modulation"] = v_mod

        orbital, o_type, o_rank, o_samb_type, _, o_lc, o_mod_type, o_mod = old["tab2"]["orbital"]
        multipie["basis"]["orbital_type"] = type_str[o_type]
        multipie["basis"]["orbital_rank"] = o_rank
        multipie["basis"]["orbital"] = orbital
        multipie["basis"]["orbital_samb_type"] = type_str[o_samb_type]
        multipie["basis"]["orbital_lc"] = o_lc
        multipie["basis"]["orbital_modulation_type"] = type_str1[o_mod_type]
        multipie["basis"]["orbital_modulation"] = o_mod

        hopping = old["tab2"]["hopping"][0]
        multipie["basis"]["hopping"] = hopping

    return multipie


# ==================================================
def get_data(dic, widget):
    """
    Get updated data.

    Args:
        dic (dict): read data.
        widget (PyVistaWidget): PyVistaWidget.

    Returns:
        - (dict) -- updated data.
    """
    widget._data["site"].block_update_widget(True)
    for row in dic["site"]:
        d = dict(zip(old_setting["site"], row))
        widget.add_site(
            0.07 * float(d["size"]), d["color"], d["opacity"], d["position"], d["cell"], d["name"], d["label"], d["space"]
        )
    widget._data["site"].block_update_widget(False)

    widget._data["bond"].block_update_widget(True)
    for row in dic["bond"]:
        d = dict(zip(old_setting["bond"], row))
        widget.add_bond(
            d["vector"],
            0.018 * float(d["width"]),
            d["color"],
            d["color2"],
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["bond"].block_update_widget(False)

    widget._data["vector"].block_update_widget(True)
    for row in dic["vector"]:
        d = dict(zip(old_setting["vector"], row))
        widget.add_vector(
            d["vector"],
            d["length"],
            0.02 * float(d["width"]),
            d["offset"],
            d["color"],
            True,
            None,
            None,
            None,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["vector"].block_update_widget(False)

    widget._data["orbital"].block_update_widget(True)
    for row in dic["orbital"]:
        d = dict(zip(old_setting["orbital"], row))
        size = -0.3 * abs(float(d["size"])) if d["scale"] else 0.3 * abs(float(d["size"]))
        widget.add_orbital(
            d["shape"],
            d["surface"],
            size,
            [[d["theta0"], d["theta1"]], [d["phi0"], d["phi1"]]],
            d["color"],
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["orbital"].block_update_widget(False)

    comp_str = {0: "x", 1: "y", 2: "z", 3: "abs"}
    widget._data["stream"].block_update_widget(True)
    for row in dic["stream"]:
        d = dict(zip(old_setting["stream"], row))
        size = -float(d["size"]) if d["scale"] else float(d["size"])
        widget.add_stream(
            d["shape"],
            d["vector"],
            size,
            [[d["theta0"], d["theta1"]], [d["phi0"], d["phi1"]]],
            [d["theta"], d["phi"]],
            d["v_size"],
            0.004 * float(d["width"]),
            None,
            to_bool(d["scale"]),
            d["color"],
            comp_str[d["component"]],
            None,
            None,
            None,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["stream"].block_update_widget(False)

    widget._data["plane"].block_update_widget(True)
    for row in dic["plane"]:
        d = dict(zip(old_setting["plane"], row))
        widget.add_plane(
            d["normal"],
            d["x"],
            d["y"],
            d["color"],
            None,
            None,
            None,
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["plane"].block_update_widget(False)

    widget._data["box"].block_update_widget(True)
    for row in dic["box"]:
        d = dict(zip(old_setting["box"], row))
        widget.add_box(
            d["a1"],
            d["a2"],
            d["a3"],
            d["width"],
            None,
            None,
            to_bool(d["wireframe"]),
            d["color"],
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["box"].block_update_widget(False)

    widget._data["polygon"].block_update_widget(True)
    for row in dic["polygon"]:
        d = dict(zip(old_setting["polygon"], row))
        widget.add_polygon(
            d["point"],
            d["connection"],
            d["width"],
            to_bool(d["edge"]),
            None,
            to_bool(d["wireframe"]),
            d["color"],
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["polygon"].block_update_widget(False)

    widget._data["text3d"].block_update_widget(True)
    for row in dic["text3d"]:
        d = dict(zip(old_setting["text3d"], row))
        widget.add_text3d(
            d["text"],
            0.04 * float(d["size"]),
            d["normal"],
            0.3 * float(d["depth"]),
            d["offset"],
            d["color"],
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["text3d"].block_update_widget(False)

    widget._data["spline"].block_update_widget(True)
    for row in dic["spline"]:
        d = dict(zip(old_setting["spline"], row))
        widget.add_spline(
            d["point"],
            0.01 * float(d["width"]),
            d["n_interp"],
            to_bool(d["closed"]),
            to_bool(d["natural"]),
            None,
            None,
            None,
            None,
            d["color"],
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["spline"].block_update_widget(False)

    widget._data["spline_t"].block_update_widget(True)
    for row in dic["spline_t"]:
        d = dict(zip(old_setting["spline_t"], row))
        widget.add_spline_t(
            d["expression"],
            d["t_range"],
            0.01 * float(d["width"]),
            d["n_interp"],
            to_bool(d["closed"]),
            to_bool(d["natural"]),
            None,
            None,
            None,
            None,
            d["color"],
            False,
            d["opacity"],
            d["position"],
            d["cell"],
            d["name"],
            d["label"],
            d["space"],
        )
    widget._data["spline_t"].block_update_widget(False)

    widget._data["caption"].block_update_widget(True)
    for row in dic["caption"]:
        d = dict(zip(old_setting["caption"], row))
        caption = "[" + ",".join(d["caption"]) + "]"
        widget.add_caption(
            caption,
            d["size"],
            to_bool(d["bold"]),
            d["color"],
            d["position"],
            d["cell"],
            d["name"],
            d["space"],
        )
    widget._data["caption"].block_update_widget(False)

    font_str = {0: "arial", 1: "courier", 2: "times"}
    widget._data["text2d"].block_update_widget(True)
    for row in dic["text"]:
        d = dict(zip(old_setting["text"], row))
        widget.add_text2d(
            d["caption"],
            d["size"],
            d["color"],
            font_str[d["font"]],
            d["position"] + [0],
            d["name"],
        )
    widget._data["text2d"].block_update_widget(False)

    data = widget.get_data_dict()

    return data


# ==================================================
def convert_version2(dic, widget):
    """
    Converter from ver.1 to ver. 2.

    Args:
        dic (dict): ver.1 dict.
        widget (PyVistaWidget): PyVistaWidget.

    Returns:
        - (dict) -- all data dict in ver. 2.
    """
    # all data.
    all_data = {
        "version": __version__,
        "data": get_data(dic, widget),
        "status": get_status(dic),
        "preference": get_preference(dic),
        "camera": get_camera(dic),
    }
    all_data["status"].update({"multipie": get_multipie(dic)})

    return all_data
